"""
The CLI for the EHub Model.

The options for running the model are set in the `config.yaml` file or as
command-line arguments. See the `config.yaml` file for more information about
the `yaml` format.

Given the config.yaml file, solving the model is easy:

    $ python run.py

The program will then print out the output of whatever solver is configured in
the config.yaml file and then the solved variables, parameters, constraints,
etc..

If you want to learn how to use `run.py` as a command-line tool, run

    $ python run.py -h

to learn more about the arguments.

Note: You can use both the `config.yaml` and the command-line arguments
together in certain situations.
"""
import argparse
import os
import pickle
import yaml

from pyehub.energy_hub.ehub_model import EHubModel
from pyehub.logger import create_logger
from pyehub.outputter import pretty_print, output_excel, print_capacities, print_section, print_warning, stream_info
from pyehub.energy_hub.utils import constraint



DEFAULT_CONFIG_FILE = 'config.yaml'
DEFAULT_LOG_FILE = 'logs.log'

# class CarbonEHubModel(EHubModel):
#     """Modified EHubModel to minimize cost with a max carbon cap."""
#     MAX_CARBON = 0

#     def __init__(self, *, excel=None, request=None, max_carbon=0, big_m=99999):
#         super().__init__(excel=excel, request=request, big_m=big_m)
#         if max_carbon is not None:
#             self.MAX_CARBON = max_carbon

#     @constraint()
#     def max_carbon_level(self):
#         """Constraint to set a max carbon cap."""
#         return self.total_carbon <= float(self.MAX_CARBON)


def main():
    """The main function for the CLI."""
    create_logger(DEFAULT_LOG_FILE)

    arguments = get_command_line_arguments()
    settings = parse_arguments(arguments)

    if settings['carbon']:
        model = EHubModel(excel=settings['input_file'], max_carbon=settings['max_carbon'], big_m=settings['big_m'])
    else:
        model = EHubModel(excel=settings['input_file'], big_m=settings['big_m'])

    # Printing out model prior to solving in the same format as after solving
    if settings['verbose']:
        print_section('Before_Solving', model._public_attributes())

    results = model.solve(settings['solver'], is_verbose=settings['verbose'])

    if not settings['quiet']:
        pretty_print(results)

    output_excel(results['solution'], settings['output_file'],
                 time_steps=len(model.time), sheets=['other', 'capacity_storage', 'capacity_tech'])

    # Outputs all the capacities form the results
    if settings['verbose']:
        print_capacities(results)

    # Prints error at the end if the model burns energy(energy from storage and energy to storage at the same time step)
    print_warning(results)

    # Outputs additional sheets for each stream with information about storages and about the stream
    if settings['output_format']:
        stream_info(results, settings['output_file'])


def get_command_line_arguments() -> dict:
    """
    Get the arguments from the command-line and return them as a dictionary.

    Returns:
        A dictionary holding all the arguments
    """
    parser = argparse.ArgumentParser()

    group = parser.add_mutually_exclusive_group()
    group.add_argument('-q', '--quiet', help='Suppresses all output',
                       action='store_true', dest='is_quiet')
    group.add_argument('-v', '--verbose', help='Output everything',
                       action='store_true', dest='is_verbose')

    parser.add_argument('--model', help='The Excel file of the model to solve',
                        dest='input_file')
    parser.add_argument('--config',
                        help=f'The path to the config file to use. Defaults '
                             f'to `{DEFAULT_CONFIG_FILE}` in the current '
                             f'directory. This is only used if the relevant '
                             f'command-line argument is not given.',
                        default=DEFAULT_CONFIG_FILE, dest='config_file')
    parser.add_argument('--output', help='The file to output the results in',
                        dest='output_file')
    parser.add_argument('--carbon', help='Add a maximum value to the carbon', action='store_true')
    parser.add_argument('--max_carbon', default=0, type=int,
                        help='The maximum value that carbon can have if the --carbon switch is used.\nDefault is 0')

    parser.add_argument('--solver_options', help='The rest of the arguments '
                                                 'are arguments to the solver',
                        action='store_true')
    parser.add_argument('--solver', help='The solver to use',
                        choices=['glpk', 'cplex'], dest='solver')

    parser.add_argument('--output_format', help='Outputs the results in a different format, grouped info about streams',
                        action='store_true')

    known, unknown = parser.parse_known_args()
    print(known, unknown)
    if not known.solver_options and unknown:
        raise ValueError('Unknown arguments. Must specify `--solver_options` '
                         'to allow extra arguments.')

    known = known.__dict__
    known['unknown'] = {}
    for key, value in zip(unknown[::2], unknown[1::2]):
        key = key.split('--')[-1]

        if key not in known['unknown']:
            known['unknown'][key] = value
        else:
            raise ValueError

    # Combine two dictionaries into one
    return known


def parse_arguments(arguments: dict) -> dict:
    """
    Parse the command-line arguments along with the config file.

    Args:
        arguments: The dictionary containing the command-line arguments

    Returns:
        A dictionary with the command-line arguments combined with the config
        file's settings
    """
    with open(arguments['config_file'], 'r') as file:
        config_settings = yaml.safe_load(file)

    if not arguments['input_file']:
        input_file = config_settings['input_file']
    elif arguments['input_file']:
        input_file = arguments['input_file']
    else:
        raise ValueError('Must specify a model to solve')

    if arguments['output_file']:
        output_file = arguments['output_file']
    else:
        output_file = config_settings['output_file']

    if arguments['solver']:
        solver_name = arguments['solver']
    else:
        solver_name = config_settings['solver']['name']

    if arguments['solver_options']:
        solver_options = arguments['unknown']
    else:
        solver_options = config_settings['solver']['options']

    big_m = config_settings['BIG_M']

    return {
        'input_file': input_file,
        'output_file': output_file,
        'verbose': arguments['is_verbose'],
        'quiet': arguments['is_quiet'],
        'solver': {
            'name': solver_name,
            'options': solver_options,
        },
        'carbon': arguments['carbon'],
        'max_carbon': arguments['max_carbon'],
        'output_format': arguments['output_format'],
        'big_m': big_m
    }


if __name__ == "__main__":
    main()



"""
Provides a function for the ehub_model using an epsilon constraint method

for input and output files - use config.yaml

To run: specify the number of intervals wanted for epsilon constraint method in the arguments

            $ python epsilon_constraint.py --epsilon_n N
"""
from energy_hub import EHubModel
from run import CarbonEHubModel
import yaml
from outputter import pretty_print, output_excel
import argparse

""" Loads a yaml file """
#TODO: yaml loader & check if file exists
with open("config.yaml", "r") as file_descriptor:
    data = yaml.safe_load(file_descriptor)


def main(epsilon_n):
    """
    Main function for running the EHub_Model using the epsilon constraint method

    :param epsilon_n: number of divisions wanted for the epsilon constraint method

    :outputs excel files with all the data (including the points in between)
    """

    outputFile = data['output_file']

    if epsilon_n is None:
        epsilon_n = 0

    # calculating the minimum cost using the EHubModel(with the maximum carbon)
    cost_min_model = EHubModel(excel=data['input_file'])
    cost_min = cost_min_model.solve(data['solver'])

    # calculating the carbon per each step of the division from the maximum carbon for epsilon_n != 0.
    carbon_per_step = 0 if epsilon_n == 0 else (cost_min['solution']['total_carbon']) / epsilon_n

    # loop that goes through all the point between the min carbon (max cost) and min cost(max carbon )
    # if epsilon_n = 0, outputs only the min carbon to the excel file
    for n in range(epsilon_n+1):

        # calculating the the points in between the two minimums, while changing the carbon constraint with
        # each iteration in the CarbonEHubModel
        carbon_min_model = CarbonEHubModel(excel=data['input_file'], max_carbon=carbon_per_step * n)
        carbon_min = carbon_min_model.solve(data['solver'])
        pretty_print(carbon_min)

        # outputs the results to separate excel files with the numbers at the end of the name from 0 - n
        # file with number 0 - displays minimizing carbon
        # file with number n - displays minimizing cost
        output_excel(carbon_min['solution'], outputFile[:-5] + "_" + str(n) + '.xlsx',
                     time_steps=len(carbon_min_model.time))


if __name__ == "__main__":

    parser = argparse.ArgumentParser()

    parser.add_argument('--epsilon_n', type=int,
                        help='The number of steps to break into for the epsilon constraint.')

    arg = parser.parse_args()

    main(arg.epsilon_n)



"""
A package that holds everything related to the request format.

This includes any wrapper classes for handling the request format.

>>> from data_formats.request_format import (
... validate, RequestValidationError, Converter, Capacity, Storage, Stream,
... TimeSeries, Network_links
... )
"""
from data_formats.request_format.request_format import (
    validate, RequestValidationError,
)
from data_formats.request_format.converter import Converter
from data_formats.request_format.capacity import Capacity
from data_formats.request_format.storage import Storage
from data_formats.request_format.stream import Stream
from data_formats.request_format.time_series import TimeSeries
from data_formats.request_format.network import Network_links



"""
A script for using the run.py module as a subprocess to solve multiple Energy Hubs while changing a given parameter.

This functionality is now also provided by the Besos EvaluatorEH.
"""

import subprocess
import openpyxl as op
import numpy as np

LETTERS = "ABCDEFGHIJKLMNOPQRSTUVWXYZ"

# Step 1: set the start and stop point of changing parameter and number of steps(+1) in line 18 to 20.
# Step 2: Set the sheet name and cell number of parameter in line 23, 24.
# Step 3: In line 22 (in_file), write the directory of the input excel.
# Step 4: In line 25 (model_out_dir), put the output directory which want to save in.
# Step 5: line 26 is the directory that the temporary file saves in. It doesn't need to change it.
# Step 6: The summary output file which contains total cost and carbon will save in line 29.
# You can change its name here.
# Step 7: in line 40, you can change the name of output file per each run.


def main():
    start = 671
    stop = 1
    steps = 11

    # in_file = 'Individual_Hubs/LI_Battery/PV_50/Diesel_27/Input_LI_PV50_D27.xlsx'
    in_file = 'Individual_Hubs/test.xlsx'
    sheet_name = 'Storages'
    cell_to_set = 'B4'
    model_out_dir = 'Individual_Hubs/LI_Battery/PV_50/Diesel_27'
    model_inputs = 'Individual_Hubs/temp_inputs.xls'
    sheet_to_read = 'Other'
    cells_to_read = ['B168', 'B170']
    final_output_file = 'Individual_Hubs/LI_Battery/PV_50/Diesel_27/output_LI_PV50_D27.xlsx'
    result_wb = op.Workbook()
    result_sheet = result_wb['Sheet']
    output_cells = ['Input value'] + cells_to_read
    for column, cell in zip(LETTERS, output_cells):
        result_sheet[f'{column}1'].value = cell

    for row,value in enumerate(np.linspace(start, stop, steps), start=2):
        wb = op.load_workbook(in_file)
        wb[sheet_name][cell_to_set].value = value
        wb.save(model_inputs)
        model_out_file = fr'{model_out_dir}/LI_PV50_D27_{int(value)}.xlsx'
        #the input file for run.py is specified in config.yaml
        print('starting subprocess')
        subprocess.run(['python', 'run.py', '--output', 'model_out_file',])
        print('file should be there')

        wb = op.load_workbook(model_out_file)
        result_sheet[f'A{row}'].value = value
        for col,cell_to_read in zip(LETTERS[1:], cells_to_read):
            print(f'writing {wb[sheet_to_read][cell_to_read].value} to {col}{row} for {value}')
            result_sheet[f'{col}{row}'].value = wb[sheet_to_read][cell_to_read].value
    result_wb.save(final_output_file)




if __name__ == '__main__':
    main()


"""
Provides functionality for handling a request format capacity.
"""
from typing import Optional


class Capacity:
    """
    A wrapper for a request format capacity.
    """

    def __init__(self, capacity_request: dict) -> None:
        """
        Create a new wrapper.

        Args:
            capacity_request: The capacity in the capacities section of the
                request format
        """
        self._capacity = capacity_request

    @property
    def name(self) -> str:
        """The name of the capacity."""
        return self._capacity['name']

    @property
    def domain(self):
        """The domain of the capacity."""
        return self._capacity['type']

    @property
    def lower_bound(self) -> Optional[float]:
        """The lower bound of the capacity as a float."""
        if 'bounds' in self._capacity and 'lower' in self._capacity['bounds']:
            return self._capacity['bounds']['lower']

        return None  # Unbounded

    @property
    def upper_bound(self) -> Optional[float]:
        """The lower bound of the capacity as a float."""
        if 'bounds' in self._capacity and 'upper' in self._capacity['bounds']:
            return self._capacity['bounds']['upper']

        return None  # Unbounded



"""
Provides a class for encapsulating an energy hub model.
"""
import inspect
import os
import sys

from itertools import product
from typing import Iterable
import logging
import yaml

import excel_to_request_format
from data_formats import response_format
from data_formats.request_format import Storage
from energy_hub.input_data import InputData
from energy_hub.param_var import ConstantOrVar
from energy_hub.utils import constraint, constraint_list
from pylp import RealVariable, IntegerVariable, BinaryVariable, Status
import pylp

DOMAIN_TO_VARIABLE = {
    'Continuous': RealVariable,
    'Integer': IntegerVariable,
    'Binary': BinaryVariable,
}

DEFAULT_SOLVER_SETTINGS = {
    'name': 'glpk',
    'options': {
        'mipgap': 0.05,
    },
}

class InfeasibleConstraintError(Exception):
    """A constraint will always be false."""
    def __init__(self, constraint_name: str = None, arguments: tuple = None) -> None:
        if arguments is None:
            message = f'Constraint {constraint_name} is False'
        else:
            message = f'Constraint {constraint_name}{arguments} is False'

        super().__init__(message)


# Allow variable docstrings
# pylint: disable=pointless-string-statement
class EHubModel:
    """
    Represents a black-box Energy Hub.
    """

    def __init__(self, *, excel=None, request=None, big_m=99999, max_carbon=None):
        """Create a new Energy Hub using some input data.

        Args:
            excel: The Excel 2003 file for input data
            request: The request format dictionary
        """
        self._data = None
        self._compiled = False
        self._objective = None
        self.BIG_M = big_m
        self.MAX_CARBON = max_carbon

        if excel:
            request = excel_to_request_format.convert(excel)

        if request:
            self._data = InputData(request)

        if self._data:
            self._constraints = []
            self._prepare()
        else:
            raise RuntimeError("Can't create a hub with no data.")

    def solve(self, solver_settings: dict = None, is_verbose: bool = False):
        """
        Solve the model.

        Args:
            solver_settings: The config options for the solver
            is_verbose: Makes it so the solver prints everything

        Returns:
            The results
        """
        if solver_settings is None:
            solver_settings = DEFAULT_SOLVER_SETTINGS

        solver = solver_settings["name"]
        options = solver_settings["options"]
        if options is None:
            options = {}

        if not self._compiled:
            self.compile()

        status = pylp.solve(objective=self.objective,
                            constraints=self.constraints,
                            minimize=True,
                            solver=solver,
                            verbose=is_verbose)

        attributes = self._public_attributes()
        return response_format.create_response(status, attributes)

    def _prepare(self):
        self._create_sets()
        self._add_parameters()

    def recompile(self):
        """Clear all constraints and variables then compile again."""
        self._constraints = []
        self._compiled = False

        self.compile()

    def compile(self):
        """Build all the constraints and variables of the model."""
        self._add_variables()
        self._add_constraints()
        self.objective = 'total_cost'

        self._compiled = True

    def _create_sets(self):
        data = self._data

        self.time = data.time
        """The time steps of the model."""
        self.technologies = data.converter_names
        """The names of the converters."""
        self.storages = data.storage_names
        """The names of the storages."""
        self.streams = data.stream_names
        """The names of all streams."""
        self.output_streams = data.output_stream_names
        """The output stream names."""
        self.demands = data.demand_stream_names
        """The demand stream names."""
        self.export_streams = data.export_streams
        """The export stream names."""
        self.import_streams = data.import_streams
        """The export stream names."""
        self.stream_timeseries = data.stream_timeseries
        """The availability time series of certain streams"""
        self.part_load = data.part_load_techs
        """The names of the converters that have a part load."""
        self.sources = data.source_stream_names
        """The sources stream names"""

    def _add_variables(self):
        self._add_capacity_variables()

        # Global variables
        self.energy_input = {

            t: {tech: RealVariable() for tech in self.technologies}
            for t in self.time
        }
        """The energy that is sent into a converter per time step."""

        self.energy_exported = {
            t: {out: RealVariable() for out in self.export_streams}
            for t in self.time
        }
        """The energy that is exported from the hub per time step."""

        self.energy_imported = {
            t: {out: RealVariable() for out in self.import_streams}
            for t in self.time
        }

        """The energy that is imported from outside the hub per time step."""

        self.capacities = ConstantOrVar(self.technologies, model=self, values=self._data.converters_capacity)

        """
        The capacity of each tech.

        The capacity could either be a constant value or it could be a
        variable.
        """

        self.is_installed = {tech: BinaryVariable()
                             for tech in self.technologies}
        """
        Is a converter installed?
        """
        self.is_installed_2 = {storage: BinaryVariable()
                               for storage in self.storages}
        """ Is storage installed? """

        self.is_on = {
            t: {tech: BinaryVariable() for tech in self.part_load}
            for t in self.time
        }
        """
        Whether a converter is turned on at a time step.

        This is only calculated for techs that have a part load limit.
        """

        self.total_cost = RealVariable()
        """The total cost of the model."""
        self.operating_cost = RealVariable()
        """The operating cost of the model."""
        self.maintenance_cost = RealVariable()
        """The maintenance cost of the model."""
        self.investment_cost = RealVariable()
        """The investment cost of the converters and storages."""

        self.total_carbon = RealVariable()
        """The total carbon produced by the model."""

        self.energy_to_storage = {
            t: {storage: RealVariable() for storage in self.storages}
            for t in self.time
        }
        """The energy sent to the storage at each time step."""
        self.energy_from_storage = {
            t: {storage: RealVariable() for storage in self.storages}
            for t in self.time
        }
        """The energy taken from a storage at each time step."""

        last_level = self.time[-1] + 1
        self.storage_level = {
            t: {storage: RealVariable() for storage in self.storages}
            for t in self.time + [last_level]
        }
        """The energy level in each storage at each point in time.

        Time steps are not points in time, but time intervals. Time step 0 is
        from time 0 up to, but not including, time 1. Time step 1 is from
        time 1 up to time 2, and so on.
        But for storages, we measure the level at each point in time. So for
        time step 0, we are measuring it at that point in time. The level
        for time step 1 is at time point 1, and so on.
        So we need an additional time "step" to measure the level after
        everything has happened in the last time step.

        ```
            time step 0   time step 1   time step 2
                 |             |             |
          |-------------|-------------|-------------|---...
        time 0        time 1        time 2        time 3
        ```
        """

        self.storage_capacity = ConstantOrVar(
            self.storages, model=self, values=self._data.storage_capacity
        )
        """
        The capacity of each storage.

        It can either be a constant value or a variable.
        """

        self.capacity_storage = {storage: self.storage_capacity[storage]
                                 for storage in self.storages}

        """Group storage capacities in a table in the output file """

        self.capacity_tech = {tech: self.capacities[tech]
                              for tech in self.technologies}

        """Group tech capacities in a table in the output file """

    def _add_parameters(self):
        data = self._data

        # coupling matrix & Technical parameters
        # coupling matrix technology efficiencies
        self.CONVERSION_EFFICIENCY = data.c_matrix
        """
        The conversion efficiency for each stream each converter.

        For example, a converter that takes in Gas and outputs Elec at 100%
        efficiency has a -1 for Gas and 1 for Elec stream.
        """

        self.MAX_CHARGE_RATE = data.storage_charge
        """
        The maximum charge rate of each storage.

        This is a constant factor of the charge coming into the storage at each
        time step.
        """
        self.MAX_DISCHARGE_RATE = data.storage_discharge
        """
        The maximum discharge rate of each storage.

        This is a constant factor of the charge coming out of the storage at
        each time step.
        """
        self.STORAGE_STANDING_LOSSES = data.storage_loss
        """
        The decay rate for each storage.

        For example, if the decay rate was 0.01 then the storage retains 99% of
        its storage level for each time step.
        """
        self.CHARGING_EFFICIENCY = data.storage_ef_ch
        """The charging efficiency for each storage."""
        self.DISCHARGING_EFFICIENCY = data.storage_ef_disch
        """The discharging efficiency for each storage."""
        self.MIN_STATE_OF_CHARGE = data.storage_min_soc
        """The minimum amount of energy the storage must have per time step."""

        self.PART_LOAD = data.part_load
        """The part load of each tech if it has it."""

        self.CARBON_FACTORS = data.carbon_factors
        self.CARBON_CREDITS = data.carbon_credits
        # Cost parameters
        # Technologies capital costs
        self.LINEAR_CAPITAL_COSTS = data.linear_cost
        self.FIXED_CAPITAL_COSTS = data.fixed_capital_costs
        self.LINEAR_STORAGE_COSTS = data.storage_lin_cost
        # Operating prices technologies
        # Maintenance costs for storages
        self.ANNUAL_MAINTENANCE_STORAGE_COSTS = data.storage_annual_maintenance_cost
        # Fixed capital costs for storages
        self.FIXED_CAPITAL_COSTS_STORAGE = data.storage_fixed_capital_cost
        self.FUEL_PRICES = data.fuel_price
        """The price of each fuel."""
        self.FEED_IN_TARIFFS = data.feed_in
        """The export price of each fuel."""
        # Maintenance costs
        self.OMV_COSTS = data.variable_maintenance_cost

        # loads
        self.LOADS = data.loads
        """The amount of energy demanded at each time step."""
        self.TIME_SERIES = data.time_series_data
        """All the time series"""

        self.NET_PRESENT_VALUE_TECH = data.tech_npv
        """The ratio for the net present value of each converter."""
        self.NET_PRESENT_VALUE_STORAGE = data.storage_npv
        """The ratio for the net present value of each storage."""


    @property
    def objective(self):
        """The objective "function" of the model."""
        # change to carbon to find the min carbon
        return self._objective

    @objective.setter
    def objective(self, new_objective):
        self._objective = self.__dict__[new_objective]


    # MAX_CARBON=0
    # @constraint()
    # def max_carbon_level(self):
    #     return self.total_carbon <= float(self.MAX_CARBON)

    @constraint('time', 'streams')
    def energy_balance(self, t, stream):
        """
        Ensure the loads and exported energy is below the produced energy.

        Args:
            t: A time step
            stream: An output stream
        """
        #logging.info(f'Stream: {stream}')

        if stream in self.demands:
            load = self.LOADS[stream][t]
        else:
            load = 0
        lhs = load

        total_q_out = 0
        total_q_in = 0
        for storage in self._get_storages_from_stream(stream):
            total_q_in += self.energy_to_storage[t][storage.name]
            total_q_out += self.energy_from_storage[t][storage.name]

        energy_output = 0
        for tech in self.technologies:  # _get_techs_from_output_stream(stream):
            # TODO: Account for techs which have two input streams
            energy_input = self.energy_input[t][tech]
            conversion_rate = self.CONVERSION_EFFICIENCY[tech][stream]
            # TODO: Account for CHP which has two output streams
            energy_output += energy_input * conversion_rate

        if stream in self.export_streams:
            energy_exported = self.energy_exported[t][stream]
        else:
            energy_exported = 0

        if stream in self.import_streams:
            energy_imported = self.energy_imported[t][stream]
        #elif stream in self.stream_timeseries:
        #    energy_imported = self.TIME_SERIES[stream][t]
        else:
            energy_imported = 0

        rhs = energy_output + total_q_out + energy_imported - total_q_in - energy_exported
        #logging.info(f'{lhs} == {rhs}')
        return lhs == rhs

    @constraint()
    def calc_total_cost(self):
        """A constraint for calculating the total cost."""
        return self.total_cost == self.investment_cost + self.operating_cost + self.maintenance_cost

    @constraint()
    def calc_total_carbon(self):
        """A constraint for calculating the total carbon produced."""
        total_carbon_credits = 0
        for stream in self.export_streams:
            # if carbon credits for a stream are a time series
            if self.CARBON_CREDITS[stream] in self.TIME_SERIES:
                for t in self.time:
                    carbon_credit = self.TIME_SERIES[self.CARBON_CREDITS[stream]][t]
                    energy_exported = self.energy_exported[t][stream]
                    total_carbon_credits += carbon_credit * energy_exported
            else:
                total_energy_exported = sum(self.energy_exported[t][stream]
                                            for t in self.time)
                carbon_credit = self.CARBON_CREDITS[stream]
                total_carbon_credits += carbon_credit * total_energy_exported

        total_carbon_emissions = 0
        for stream in self.import_streams:
            # if carbon factors for a stream are a time series
            if self.CARBON_FACTORS[stream] in self.TIME_SERIES:
                for t in self.time:
                    carbon_factor = self.TIME_SERIES[self.CARBON_FACTORS[stream]][t]
                    energy_used = self.energy_imported[t][stream]
                    total_carbon_emissions += carbon_factor * energy_used
            else:
                total_energy_used = sum(self.energy_imported[t][stream]
                                        for t in self.time)
                carbon_factor = self.CARBON_FACTORS[stream]
                total_carbon_emissions += carbon_factor * total_energy_used

        return self.total_carbon == total_carbon_emissions - total_carbon_credits

    @constraint()
    def calc_investment_cost(self):
        """A constraint for calculating the investment cost."""
        storage_cost = sum(self.NET_PRESENT_VALUE_STORAGE[storage]
                           * self.LINEAR_STORAGE_COSTS[storage]
                           * self.storage_capacity[storage]
                           + self.FIXED_CAPITAL_COSTS_STORAGE[storage]
                           * self.is_installed_2[storage]
                           for storage in self.storages)

        tech_cost = sum(self.NET_PRESENT_VALUE_TECH[tech]
                        * (self.LINEAR_CAPITAL_COSTS[tech]
                           * self.capacities[tech]
                           + self.FIXED_CAPITAL_COSTS[tech]
                           * self.is_installed[tech])
                        for tech in self.technologies)

        cost = tech_cost + storage_cost
        return self.investment_cost == cost

    @constraint()
    def calc_maintenance_cost(self):
        """A constraint for calculating the maintenance cost."""
        cost = 0
        for t in self.time:
            for tech in self.technologies:
                for energy in self.output_streams:
                    cost += (self.energy_input[t][tech]
                             * self.CONVERSION_EFFICIENCY[tech][energy]
                             * self.OMV_COSTS[tech])

        for storage in self.storages:
            cost += self.ANNUAL_MAINTENANCE_STORAGE_COSTS[storage] * self.is_installed_2[storage]

        return self.maintenance_cost == cost

    @constraint()
    def calc_operating_cost(self):
        """A constraint for calculating the total operating cost."""
        total_export_income = 0
        for stream in self.export_streams:
            # If the price is time series
            if self.FEED_IN_TARIFFS[stream] in self.TIME_SERIES:
                for t in self.time:
                    price = self.TIME_SERIES[self.FEED_IN_TARIFFS[stream]][t]
                    energy_exported = self.energy_exported[t][stream]
                    total_export_income += price * energy_exported

            else:
                total_energy_exported = sum(self.energy_exported[time][stream]
                                            for time in self.time)
                price = self.FEED_IN_TARIFFS[stream]
                total_export_income += price * total_energy_exported


        total_fuel_bill = 0
        for stream in self.import_streams:
            # If the price is a time series
            if self.FUEL_PRICES[stream] in self.TIME_SERIES:
                for t in self.time:
                    price = self.TIME_SERIES[self.FUEL_PRICES[stream]][t]
                    energy_used = self.energy_imported[t][stream]
                    total_fuel_bill += price * energy_used

            else:
                total_energy_used = sum(self.energy_imported[time][stream]
                                        for time in self.time)
                price = self.FUEL_PRICES[stream]
                total_fuel_bill += price * total_energy_used

        return self.operating_cost == total_fuel_bill - total_export_income

    @constraint('technologies')
    def tech_is_installed(self, tech):
        """
        Set binary to 1 if capacity of tech is > 0.
        Args:
            tech: A converter
        """
        capacity = self.capacities[tech]
        rhs = self.BIG_M * self.is_installed[tech]
        #logging.info(f'{capacity} <= {rhs}')
        return capacity <= rhs

    @constraint('technologies')
    def tech_is_installed_2(self, tech):
        """
        Set binary to 1 if capacity of tech is > 0.
        Args:
            tech: A converter
        """
        installed = self.is_installed[tech]
        lhs = self.BIG_M * self.capacities[tech]
        #logging.info(f'{lhs} >= {installed}')
        return lhs >= installed

    @constraint('storages')
    def storage_is_installed(self, storage):
        """
        Set binary to 1 if capacity of storage is > 0.
        Args:
            storage: A storage
        """
        capacity = self.storage_capacity[storage]
        rhs = self.BIG_M * self.is_installed_2[storage]
        return capacity <= rhs

    @constraint('storages')
    def storage_is_installed_2(self, storage):
        """
        Set binary to 1 if capacity of storage is > 0.
        Args:
            storage: A storage
        """
        installed = self.is_installed_2[storage]
        lhs = self.BIG_M * self.storage_capacity[storage]
        return lhs >= installed

    @constraint('time', 'part_load', 'output_streams')
    def tech_is_on(self, t, disp, out):
        """
        Set binary to on if tech is active.

        Args:
            t: A time step
            disp: A dispatch tech
            out: An output energy stream
        """
        energy_input = self.energy_input[t][disp]
        is_on = self.is_on[t][disp]

        lhs = energy_input #* conversion_rate
        rhs = self.BIG_M * is_on
        logging.info(f'{lhs} <= {rhs}')
        return lhs <= rhs

    @constraint('time', 'part_load', 'output_streams')
    def tech_is_on_2(self, t, disp, out):
        """
        Set binary to on if tech is active.

        Args:
            t: A time step
            disp: A dispatch tech
            out: An output energy stream
        """
        energy_input = self.energy_input[t][disp]
        is_on = self.is_on[t][disp]

        lhs = self.BIG_M * energy_input #* conversion_rate
        rhs = is_on

        return lhs >= rhs

    @constraint('time', 'part_load', 'output_streams')
    def tech_part_loads(self, t, disp, out):
        """
        Args:
            t: A time step
            disp: A dispatch tech
            out: An output energy stream
        """
        conversion_rate = self.CONVERSION_EFFICIENCY[disp][out]

        if conversion_rate <= 0:
            return None

        part_load = self.PART_LOAD[disp][out]
        capacity = self.capacities[disp]
        energy_input = self.energy_input[t][disp]
        is_on = self.is_on[t][disp]

        lhs = part_load * capacity

        rhs = (energy_input * conversion_rate
               + self.BIG_M * (1 - is_on))
        return lhs <= rhs

    # TODO: Separate out tech-level energy balance - define over tech and streams?
    # @constraint('time', 'technologies')
    # def tech_balance(self, t, tech):
    #     """
    #     Energy output from a converter is the input time the efficiency.
    #
    #     Args:
    #         t: A time step
    #         tech: A converter technology
    #     """
    #     energy_output = self.energy_output[t][tech]
    #     for stream in self.output_streams:
    #         energy_input = self.energy_input[t][tech]
    #         conversion_rate = self.CONVERSION_EFFICIENCY[tech][stream]
    #
    #     return energy_output == energy_input * conversion_rate

    @constraint('time', 'technologies')
    def tech_input_below_capacity(self, t, tech):
        """
        Ensure the energy input by a tech is less than its capacity.

        Args:
            t: A time step
            tech: A converter
        """
        energy_input = self.energy_input[t][tech]
        capacity = self.capacities[tech]

        techinput = next(self._get_tech_input(tech)) # Get name of input stream

        if techinput[0] in self.sources:
            timecap = self.TIME_SERIES[techinput[0]][t]

            return energy_input == timecap * capacity
        else:
            return energy_input <= capacity

    @constraint('time', 'technologies')
    def tech_input_positive(self, t, tech):
        """Energy input to tech should be positive."""
        return self.energy_input[t][tech] >= 0

    @constraint('time', 'export_streams')
    def tech_export_positive(self, t, stream):
        """Energy exported should be positive."""
        return self.energy_exported[t][stream] >= 0

    @constraint('time', 'import_streams')
    def tech_import_positive(self, t, stream):
        """Energy exported should be positive."""
        return self.energy_imported[t][stream] >= 0

    @constraint('time', 'storages')
    def storage_level_below_capacity(self, t, storage):
        """
        Ensure the storage's level is below the storage's capacity.

        Args:
            t: A time step
            storage: A storage
        """
        storage_level = self.storage_level[t][storage]
        storage_capacity = self.storage_capacity[storage]

        return storage_level <= storage_capacity

    @constraint('time', 'storages')
    def storage_level_above_minimum(self, t, storage):
        """
        Ensure the storage level is above it's minimum level.

        Args:
            t: A time step
            storage: A storage
        """
        storage_capacity = self.storage_capacity[storage]
        min_soc = self.MIN_STATE_OF_CHARGE[storage]
        storage_level = self.storage_level[t][storage]

        min_storage_level = storage_capacity * min_soc

        return min_storage_level <= storage_level

    @constraint('time', 'storages')
    def storage_discharge_rate(self, t, storage):
        """
        Ensure the discharge rate of a storage is below it's maximum rate.

        Args:
            t: A time step
            storage: A storage
        """
        max_discharge_rate = self.MAX_DISCHARGE_RATE[storage]
        storage_capacity = self.storage_capacity[storage]
        discharge_rate = self.energy_from_storage[t][storage]

        max_rate = max_discharge_rate * storage_capacity

        return discharge_rate <= max_rate

    @constraint('time', 'storages')
    def storage_charge_rate(self, t, storage):
        """
        Ensure the charge rate of a storage is below it's maximum rate.

        Args:
            t: A time step
            storage: A storage
        """
        max_charge_rate = self.MAX_CHARGE_RATE[storage]
        storage_capacity = self.storage_capacity[storage]
        charge_rate = self.energy_to_storage[t][storage]

        max_rate = max_charge_rate * storage_capacity

        return charge_rate <= max_rate

    @constraint('time', 'storages')
    def storage_balance(self, t, storage):
        """
        Calculate the current storage level from the previous level.

        Args:
            t: A time step
            storage: A storage
        """
        # See the storage_level declaration for more details
        next_storage_level = self.storage_level[t + 1][storage]
        current_storage_level = self.storage_level[t][storage]

        storage_standing_loss = self.STORAGE_STANDING_LOSSES[storage]

        discharge_rate = self.DISCHARGING_EFFICIENCY[storage]
        charge_rate = self.CHARGING_EFFICIENCY[storage]

        charge_in = self.energy_to_storage[t][storage]
        charge_out = self.energy_from_storage[t][storage]

        calculated_next_storage_level = (
            ((1 - storage_standing_loss) * current_storage_level)
            + (charge_rate * charge_in)
            - ((1 / discharge_rate) * charge_out)
        )
        return next_storage_level == calculated_next_storage_level

    @constraint('time', 'storages')
    def storage_input_positive(self, t, storage):
        """Energy to storage should be positive."""
        return self.energy_to_storage[t][storage] >= 0

    @constraint('time', 'storages')
    def storage_output_positive(self, t, storage):
        """Energy from the storages should be positive."""
        return self.energy_from_storage[t][storage] >= 0

    @constraint('time', 'storages')
    def storage_level_positive(self, t, storage):
        """Storages' levels should be above zero."""
        return self.storage_level[t][storage] >= 0

    @constraint('storages')
    def storage_looping(self, storage):
        """Ensure that the storage level at the beginning is equal to it's end
        level."""

         # See the storage_level declaration
        last_entry = list(self.time)[-1] + 1
        first_entry = list(self.time)[0]

        start_level = self.storage_level[first_entry][storage]
        end_level = self.storage_level[last_entry][storage]

        return start_level == end_level

    @constraint_list()
    def capacity_bounds(self):
        """Ensure the capacities are within their given bounds."""
        for capacity in self._data.capacities:
            variable = getattr(self, capacity.name)

            lower_bound = capacity.lower_bound
            upper_bound = capacity.upper_bound

            if lower_bound is not None and upper_bound is not None:
                yield lower_bound <= variable
                yield variable <= upper_bound
            elif lower_bound is not None and upper_bound is None:
                yield lower_bound <= variable
            elif lower_bound is None and upper_bound is not None:
                yield variable <= upper_bound
            else:
                raise RuntimeError

    @constraint()
    def max_carbon_level(self):
        """Constraint to set a max carbon cap."""
        if self.MAX_CARBON is not None:
            return self.total_carbon <= float(self.MAX_CARBON)


    # *****************************************************************************************************************
    # EVERYTHING BELOW HERE IS A HELPER FUNCTION, NOT PART OF THE MODEL
    # *****************************************************************************************************************

    def _get_storages_from_stream(self, out: str) -> Iterable[Storage]:
        return (storage for storage in self._data.storages
                if storage.stream == out)

    def _get_techs_from_output_stream(self, out: str):
        return (tech.name for tech in self._data.converters
                if tech.outputs == out)

    def _get_tech_input(self, techname: str):
        for tech in self._data.converters:
            if tech.name == techname:
                yield tech.inputs

    def _add_constraint_to_constraints(self, constraint_):
        # Comparing to True and False is bad, but expression can
        # be many types
        if constraint_ is True:
            return
        elif constraint_ is False:
            raise InfeasibleConstraintError()
        elif constraint_ is not None:
            self.constraints.append(constraint_)

    def _add_indexed_constraints(self):
        """Add all the constraint decorated functions to the model."""
        methods = [getattr(self, method)
                   for method in dir(self)
                   if callable(getattr(self, method))]
        rules = (rule for rule in methods if hasattr(rule, 'is_constraint'))

        for rule in rules:
            if not rule.enabled:
                logging.info(f'Constraint: {rule.__name__} NOT ENABLED')
                continue

            logging.info(f'Constraint: {rule.__name__}')
            if not rule.args:
                # Constraint is not indexed by anything
                expression = rule()

                try:
                    self._add_constraint_to_constraints(expression)
                except InfeasibleConstraintError:
                    raise InfeasibleConstraintError(rule.__name__)
            else:
                sets = [getattr(self, arg) for arg in rule.args]

                for indices in product(*sets):
                    expression = rule(*indices)

                    try:
                        self._add_constraint_to_constraints(expression)
                    except InfeasibleConstraintError:
                        raise InfeasibleConstraintError(rule.__name__, indices)

    def _add_constraint_lists(self):
        methods = [getattr(self, method)
                   for method in dir(self)
                   if callable(getattr(self, method))]
        rules = (rule for rule in methods
                 if hasattr(rule, 'is_constraint_list'))

        for rule in rules:
            if not rule.enabled:
                logging.info(f'Constraint List: {rule.__name__} NOT ENABLED')
                continue

            logging.info(f'Constraint List: {rule.__name__}')

            for expression in rule():
                self.constraints.append(expression)

    def _add_constraints(self):
        self._add_indexed_constraints()
        self._add_constraint_lists()

    def _add_capacity_variables(self):
        for capacity in self._data.capacities:
            domain = capacity.domain
            name = capacity.name

            try:
                variable = DOMAIN_TO_VARIABLE[domain]()
            except KeyError:
                raise ValueError(
                    f'Cannot create variable of type: {domain}. Can only be: '
                    f'{list(DOMAIN_TO_VARIABLE.keys())}'
                )

            setattr(self, name, variable)

    @property
    def constraints(self):
        """The list of constraints on the model."""
        return self._constraints

    @constraints.setter
    def constraints(self, constraints):
        """Set the constraints of the model."""
        self._constraints = constraints

    @property
    def solution_dict(self):
        fake_status = Status(status='', time=0)
        attributes = self._public_attributes()

        response = response_format.create_response(fake_status, attributes)

        return response['solution']

    def _public_attributes(self):
        return {name: value
                for name, value in self.__dict__.items()
                if not name.startswith('_')}



import os
import csv
from pyehub.outputter import pretty_print
from pyehub.energy_hub.ehub_model import EHubModel
from pyehub.energy_hub.utils import constraint, constraint_list
from pprint import pprint as pp

#TODO: Replace in the main model the data.carbon_factors with the actual time resolved time series from the SILVER CSV
# self.CARBON_FACTORS = data.carbon_factors

with open('hourly_GHG_emission.csv') as csvfile:
    readCSV = csv.reader(csvfile, delimiter=',')
    count = 0
    gen_list = []
    date_list = []
    carbon_factors = {}
    for row in readCSV:
        length = len(row)
        for column in range(length):
           #Note each generator
            if(count == 0):
                if(column!=0):
                    gen_list.append(row[column])
            #Make a key for each timeseries
            else:
                if(column==0):
                    date_list.append(row[column])
                    carbon_factors[row[column]] = {}
                    for key in gen_list:
                        carbon_factors[row[column]][key]= {}
                else:
                    carbon_factors[date_list[count-1]][gen_list[column-1]]=row[column]

        count += 1
        
carbon_factors.pop('kind')
#     pp(carbon_factors)
# pp(carbon_factors[gen_list[63]][date_list[-1]])
# pp(gen_list)
# print(length)
#     pp(date_list)
pp(carbon_factors)
    

class TimeResolvedCarbon(EHubModel):
    """
    A custom EHubModel to include the time resolved carbon factors from the Silver model
    instead of using the averaged carbon factors for each input stream.
    """
    #TODO: Override the calc_total_carbon constraint to use the time series instead. 
    #(currently I've added the default here.)
    @constraint()
    def calc_total_carbon(self):
        """A constraint for calculating the total carbon produced."""
        total_carbon_credits = 0
        for stream in self.export_streams:
            # if carbon credits for a stream are a time series
            if self.CARBON_CREDITS[stream] in self.TIME_SERIES:
                for t in self.time:
                    carbon_credit = self.TIME_SERIES[self.CARBON_CREDITS[stream]][t]
                    energy_exported = self.energy_exported[t][stream]
                    total_carbon_credits += carbon_credit * energy_exported
            else:
                total_energy_exported = sum(self.energy_exported[t][stream]
                                            for t in self.time)
                carbon_credit = self.CARBON_CREDITS[stream]
                total_carbon_credits += carbon_credit * total_energy_exported

        total_carbon_emissions = 0
        for stream in self.import_streams:
            # if carbon factors for a stream are a time series
            if self.CARBON_FACTORS[stream] in self.TIME_SERIES:
                for t in self.time:
                    carbon_factor = self.TIME_SERIES[self.CARBON_FACTORS[stream]][t]
                    energy_used = self.energy_imported[t][stream]
                    total_carbon_emissions += carbon_factor * energy_used
            else:
                total_energy_used = sum(self.energy_imported[t][stream]
                                        for t in self.time)
                carbon_factor = self.CARBON_FACTORS[stream]
                total_carbon_emissions += carbon_factor * total_energy_used

        return self.total_carbon == total_carbon_emissions - total_carbon_credits




"""
Provides functionality for handling a request format's converter.
"""
from typing import List, Optional, Union, Dict


class Converter:
    """A wrapper for a request format converter."""

    def __init__(self, converter_request: dict, capacity_converter: dict) -> None:
        """Create a new wrapper for a converter.

        Args:
            converter_request: The converter in the request format
            capacity_converter: The capacity associated with the converter
        """
        self._converter = converter_request
        self._capacity = capacity_converter

    @property
    def capacity(self) -> Union[float, str]:
        """Return the capacity of the converter."""
        return self._converter['capacity']

    @property
    def min_load(self) -> Optional[float]:
        """Return the minimum load of the tech if it has one."""
        if 'min_load' in self._converter:
            return self._converter['min_load'] / 100
        return None

    @property
    def is_solar(self) -> bool:
        """Is the converter solar?"""
        return self.is_roof_tech

    @property
    def max_capacity(self) -> float:
        """The maximum capacity of the converter."""
        return self.capacity

    @property
    def has_part_load(self) -> bool:
        """Does the converter have a part load?"""
        if 'min_load' in self._converter:
            part_load = self._converter['min_load']

            return part_load > 0

        return False

    @property
    def inputs(self) -> List[str]:
        """The names of the input streams for the converter."""
        return self._converter['inputs']

    @property
    def name(self) -> str:
        """The name of the converter."""
        return self._converter['name']

    @property
    def outputs(self) -> List[str]:
        """The names of the output streams for the converter."""
        return self._converter['outputs']

    @property
    def efficiency(self) -> float:
        """The efficiency of the converter."""
        return self._converter['efficiency']

    @property
    def output_ratios(self) -> Dict[str, float]:
        """The output ratios for each output stream."""
        output_ratios = {}
        for output in self.outputs:
            if output == 'Elec':
                output_ratios['Elec'] = 1
            else:
                output_ratios[output] = self._output_ratio

        return output_ratios

    @property
    def _output_ratio(self) -> float:
        """The output efficiency of the converter."""
        if 'output_ratio' in self._converter:
            return self._converter['output_ratio']

        return 1.0

    @property
    def usage_maintenance_cost(self) -> float:
        """The usage maintenance cost of the converter."""
        return self._converter['usage_maintenance_cost']

    @property
    def lifetime(self) -> float:
        """The lifetime in years of the tech."""
        return self._converter['lifetime']

    @property
    def is_roof_tech(self) -> bool:
        """Is this converter on the roof?"""
        return 'Irradiation' in self.inputs

    @property
    def is_dispatch(self) -> bool:
        """Is this a dispatch converter?"""
        return (not self.is_solar
                and self.name != 'Grid')

    @property
    def fixed_capital_cost(self) -> float:
        """
        Return the fixed capital cost of the converter.

        Returns:
            The fixed capital cost if it has one or 0.
        """
        if 'fixed_capital_cost' in self._converter:
            return self._converter['fixed_capital_cost']

        return 0

    @property
    def capital_cost(self) -> float:
        """Return the capital cost of the converter."""
        return self._converter['capital_cost']

    @property
    def is_grid(self) -> bool:
        """Is this converter the grid?"""
        return self._converter['name'] == 'Grid'

    @property
    def area(self) -> Optional[float]:
        """Return the area of this roof tech if this tech is a roof tech."""
        if self._capacity and self._capacity['units'] == 'm2':
            return self._capacity['bounds']['lower']

        return None

    @property
    def is_chp(self) -> bool:
        """Does the converter use and/or output heat and power?"""
        return len(self._converter['outputs']) >= 2



"""
Provides some utility functions for all code in energy_hub.
"""
import functools


def cached_property(func):
    """Return a property that caches results.

    Args:
        func: The function to decorated

    Returns:
        The decorated cached property
    """
    @property
    @functools.lru_cache(maxsize=1)
    @functools.wraps(func)
    def _wrapper(*args, **kwargs):
        return func(*args, **kwargs)

    return _wrapper


def constraint(*args, enabled=True):
    """
    Mark a function as a constraint of the model.

    The function that adds these constraints to the model is
    `_add_indexed_constraints`.

    Args:
        *args: The indices that the constraint is indexed by. Each element of
            each index is passed to the method.
        enabled: Is the constraint enabled? Defaults to True.

    Returns:
        The decorated function
    """
    def _wrapper(func):
        functools.wraps(func)

        func.is_constraint = True
        func.args = args
        func.enabled = enabled

        return func

    return _wrapper


def constraint_list(*, enabled=True):
    """
    Mark a function as a ConstraintList of the model.

    The function has to return a generator. ie: must use a yield in the method
    body.

    Args:
        enabled: Is the constraint enabled? Defaults to True.

    Returns:
        The decorated function
    """
    def _wrapper(func):
        functools.wraps(func)

        func.is_constraint_list = True
        func.enabled = enabled

        return func

    return _wrapper



"""
Contains functionality for dealing with expressions.

Exports:
    >>> from pylp.expression import Expression
"""
from contextlib import suppress
from typing import Iterable, List

from pylp.constraint import Constraint


class Expression:
    """
    Represents an expression in a linear programming problem.

    Notes:
        A list of operands is given in order to reduce the height of the
        expression tree.

        Normally, an operator operates on two operands. Thus the expression:
            5 + 5 + 5
        can be represented as a tree:
            +
           /\
          5  +
            / \
           5   5

        But as the expression gets longer, a naive approach would result in
        a very tall tree. In order to evaluate the tree, we would have to
        traverse all the way to its leafs, which at its deepest part would
        probably result in a stack overflow (maximum recursion error).

        But treating the operands as a list, results in a nicer tree:
           +
         / | \
        5  5  5

        And this tree stays at the same height the more operands it holds.

        This prevents a stack overflow from occurring. This also results in
        better performance as well.
    """

    def __init__(self, operator: str, operands: list) -> None:
        """
        Create a new expression for the given operator and operands.

        Args:
            operator: The str of the operator
            operands: A list of operands
        """
        self.operator = operator

        self.arguments = []
        self._set_arguments(operands)

    def _set_arguments(self, arguments: list) -> None:
        for arg in arguments:
            if isinstance(arg, Expression) and self.is_same_type(arg):
                self.arguments += arg.arguments
            else:
                self.arguments += [arg]

    def is_same_type(self, other: 'Expression') -> bool:
        """Return True if the other expression is of the same type."""
        return self.operator == other.operator

    def __str__(self) -> str:
        arguments = (str(arg) for arg in self.arguments)

        expression = f' {self.operator} '.join(arguments)

        return f'({expression})'

    def __le__(self, other) -> Constraint:
        return Constraint(self, '<=', other)

    def __ge__(self, other) -> Constraint:
        return Constraint(self, '>=', other)

    def __eq__(self, other) -> Constraint:
        return Constraint(self, '==', other)

    def __ne__(self, other) -> Constraint:
        raise TypeError('Cannot have != in linear programming.')

    def __lt__(self, other) -> Constraint:
        return Constraint(self, '<', other)

    def __gt__(self, other) -> Constraint:
        return Constraint(self, '>', other)

    def __sub__(self, other) -> 'Expression':
        return Expression('-', [self, other])

    def __rsub__(self, other) -> 'Expression':
        return Expression('-', [other, self])

    def __add__(self, other) -> 'Expression':
        return Expression('+', [self, other])

    def __radd__(self, other) -> 'Expression':
        return Expression('+', [other, self])

    def __mul__(self, other) -> 'Expression':
        return Expression('*', [self, other])

    def __rmul__(self, other) -> 'Expression':
        return Expression('*', [other, self])

    def __neg__(self) -> 'Expression':
        return Expression('*', [-1, self])

    def __truediv__(self, other) -> 'Expression':
        return Expression('*', [self, 1 / other])

    def __rtruediv__(self, other):
        # An expression will always have a variable in it. Otherwise, it would
        # use the regular int's or float's operations.
        raise TypeError('Cannot divide by a variable in linear programming.')

    def _handle(self, arguments):
        # Lazy evaluate to prevent errors and reduce computation
        return {
            '+': lambda: self._handle_add(arguments),
            '*': lambda: self._handle_multiplication(arguments),
            '-': lambda: self._handle_subtraction(arguments),
        }[self.operator]()

    @staticmethod
    def _handle_add(arguments: Iterable['Expression']):
        result = 0
        for arg in arguments:
            result += arg

        return result

    @staticmethod
    def _handle_subtraction(arguments: List['Expression']):
        result = arguments[0]
        for arg in arguments[1:]:
            result -= arg

        return result

    @staticmethod
    def _handle_multiplication(arguments: Iterable['Expression']):
        result = 1
        for arg in arguments:
            result *= arg

        return result

    def evaluate(self):
        """Evaluate the expression."""
        arguments = []
        for arg in self.arguments:
            with suppress(AttributeError):
                arg = arg.evaluate()
            arguments.append(arg)

        try:
            return self._handle(arguments)
        except KeyError:
            raise ValueError(f'Using unsupported operator: {self.operator}')

    def construct(self):
        """Build the expression for use in the solver."""
        arguments = []
        for arg in self.arguments:
            with suppress(AttributeError):
                arg = arg.construct()
            arguments.append(arg)

        try:
            return self._handle(arguments)
        except KeyError:
            raise ValueError(f'Using unsupported operator: {self.operator}')



"""
Provides a class that can either be a reference to a value or to a variable.
"""
from typing import List, Dict


class ConstantOrVar:
    """
    Provides access to data that can either be a constant or a variable.

    The values can either be a constant (like a float) or can be a str, which
    is a reference to a variable in the model.
    """

    def __init__(self, *indices: List, model=None,
                 values: Dict = None) -> None:
        """
        Create a new class.

        Args:
            *indices: The indices of which to index by
            model: The model that holds the variables
            values: The dictionary to initialize the object. The keys are
                indexed by *indices and their values are either constants or
                strings that reference a variable.
        """
        if model is None or values is None:
            raise ValueError
        self._model = model
        self._indices = indices
        self._values = values

    def __getitem__(self, item):
        value = self._values[item]
        if isinstance(value, str):
            # References a variable
            return getattr(self._model, value)

        return value

    def _get_values(self, matrix: Dict) -> Dict:
        for key, value in matrix.items():
            if isinstance(value, dict):
                value = self._get_values(value)
            elif hasattr(value, 'value'):
                value = value.name

            matrix[key] = value

        return matrix

    @property
    def values(self):
        """Return the values."""
        return self._get_values(self._values)



"""
A script that converts a network Excel file into the request format.

The Excel file has to be in one of the supported formats, which examples can
be found in the data_formats directory.
"""
import xlrd
from contextlib import suppress


def convert(excel_file):
    for subclass in Converter.__subclasses__():
        converter = subclass(excel_file)

        return converter.convert()


class Converter:
    def __init__(self, excel_file):
        self._excel_file = excel_file
        self._file = xlrd.open_workbook(excel_file)

    def _get_columns(self, sheet_name, start=0):
        sheet = self._file.sheet_by_name(sheet_name)

        for colx in range(start, sheet.ncols):
            yield sheet.col_values(colx)[1:] # slicing the first row

    def _get_capacities(self):
        raise NotImplementedError
    def _get_links(self):
        raise NotImplementedError
    def _get_network(self):
        raise NotImplementedError

    def convert(self):
        network_request = {
            'capacities': self._get_capacities(),
            'links': self._get_links(),
            'network': self._get_network()
        }

        return network_request


class NewFormatConverter(Converter):
    def _get_columns(self, sheet_name, start=0):
        return super()._get_columns(sheet_name, start=1)

    def _get_capacities(self):
        capacities = []
        for column in self._get_columns('Capacities'):
            (name, units, item_type, options, lower_bound, upper_bound) = column

            if name not in ['#', '']: # to get ipysheet to work
                capacity = {
                    'name': name,
                    'units': units,
                    'type': item_type
                }

                if lower_bound != '' or upper_bound != '':
                    capacity['bounds'] = {}

                    with suppress(ValueError):
                        capacity['bounds']['lower'] = int(lower_bound)
                    with suppress(ValueError):
                        capacity['bounds']['upper'] = int(upper_bound)

                capacities.append(capacity)

        return capacities

    def _get_links(self):
        links = []
        for column in self._get_columns('Network links'):
            (link_id, start_id, end_id, link_type, length, capacity, voltage, electrical_resistance,
             electrical_reactance, total_thermal_loss, total_pressure_loss, operating_temperature) = column

            if link_id not in ['#', '']: # to get ipysheet to work
                with suppress(ValueError):
                    link_id = int(link_id)
                with suppress(ValueError):
                    start_id = int(start_id)
                with suppress(ValueError):
                    end_id = int(end_id)
                with suppress(ValueError):
                    link_type = str(link_type)
                with suppress(ValueError):
                    length = float(length)
                with suppress(ValueError):
                    voltage = float(voltage)
                with suppress(ValueError):
                    electrical_resistance = float(electrical_resistance)
                with suppress(ValueError):
                    electrical_reactance = float(electrical_reactance)
                with suppress(ValueError):
                    total_thermal_loss = float(total_thermal_loss)
                with suppress(ValueError):
                    total_pressure_loss = float(total_pressure_loss)
                with suppress(ValueError):
                    operating_temperature = float(operating_temperature)

                link ={
                    'id': link_id,
                    'start_id': start_id,
                    'end_id': end_id,
                    'type': link_type,
                    # 'link_proportional_cost': link_proportional_cost, uncomment if want to set link cost in the excel file.
                    # add 'link_cost' in the column names
                    'length': length,
                    'voltage': voltage,
                    'resistance': electrical_resistance,
                    'reactance': electrical_reactance,
                    'total_thermal_loss': total_thermal_loss,
                    'total_pressure_loss': total_pressure_loss,
                    'operating_temp': operating_temperature,
                }

                try:
                    capacity = float(capacity)
                except ValueError:
                    capacity = str(capacity)

                link['capacity'] = capacity

                links.append(link)

        return links

    def _get_network(self):
        sheet = self._file.sheet_by_name('Network')

        return{
            'fixed_network_investment_cost': float(sheet.cell(1, 1).value),
            'link_proportional_cost': float(sheet.cell(2, 1).value), } # shifting the sheets one row below required to change the row indices here








"""
A script for printing out the contents of an excel spreadsheet.
"""

import openpyxl
from openpyxl import load_workbook

def main():
    """Main function."""
    book = openpyxl.load_workbook('Individual_Hubs/LI_Batresuls_11.xlsx')
    print(book.get_sheet_names())
    sheet = book.get_sheet_by_name('Other')

    # No of written Rows in sheet
    num_row = sheet.max_row
    # No of written Columns in sheet
    num_column = sheet.max_column
    # Reading each cell in excel
    for i in range(1, num_row+1):
        for j in range(1, num_column+1):
            print(sheet.cell(row=i, column=j).value)

if __name__ == '__main__':
    main()



"""
Contains functionality for dealing with a linear programming model.
"""
from typing import Iterable
from collections import namedtuple

import pulp
from contexttimer import Timer

from pylp.constraint import Constraint
import warnings


if pulp.__version__ == '2.1':
    import pulp.apis.core as solvers
    import pulp.apis.cplex_api as cplex
    import pulp.apis.glpk_api as glpk
else:
    import pulp.solvers as solvers
    warnings.warn('You are using pulp 2.0 or lower, pulp.apis.core has been changed to pulp.sovers automatically')

Status = namedtuple('Status', ['status', 'time'])


def solve(*, objective=None, constraints: Iterable[Constraint] = None,
          minimize: bool = False, solver: str = 'glpk',
          verbose: bool = False) -> Status:
    """
    Solve the linear programming problem.

    Args:
        objective: The objective function
        constraints: The collection of constraints
        minimize: True for minimizing; False for maximizing
        solver: The solver to use. Current supports 'glpk' and 'cplex'.
        verbose: If True, output the results of the solver

    Returns:
        A tuple of the status (eg: Optimal, Unbounded, etc.) and the elapsed
        time
    """
    if minimize:
        sense = pulp.LpMinimize
    else:
        sense = pulp.LpMaximize

    problem = pulp.LpProblem(sense=sense)

    # Objective function is added first
    problem += objective.construct()

    if constraints:
        for constraint in constraints:
            problem += constraint.construct()

    if solver == 'glpk' and pulp.__version__ == '2.1':
        solver = glpk.GLPK_CMD(msg=verbose)
    elif solver == 'glpk' and pulp.__version__ != '2.1':
        solver = solvers.GLPK(msg=verbose)
    elif solver == 'cplex' and pulp.__version__ == '2.1':
        solver = cplex.CPLEX_CMD(msg=verbose)
    elif solver == 'cplex' and pulp.__version__ != '2.1':
        solver = solvers.CPLEX(msg=verbose)
    else:
        raise ValueError(f'Unsupported solver: {solver}')

    with Timer() as time:
        results = problem.solve(solver)

    status = pulp.LpStatus[results]

    return Status(status=status, time=time.elapsed)



"""
This script is an interface for using the logging module to log the running of energy hub model.
"""
import logging

def create_logger(filename: str) -> None:
    """
    Add logging to the application.

    Args:
        filename: The name of the log file
    """
    # logging accepts all logs, which are then handled later
    logging.getLogger().setLevel(logging.DEBUG)

    add_console_handler()
    add_file_handler(filename)


def add_console_handler() -> None:
    """Add a logging handler for stderr."""
    # stderr
    console_handler = logging.StreamHandler()
    console_handler.setLevel(logging.WARNING)
    console_handler.setFormatter(get_default_formatter())

    logging.getLogger().addHandler(console_handler)


def add_file_handler(filename: str) -> None:
    """
    Add a logging handler for a log file.

    Args:
        filename: The name of the log file
    """
    file_handler = logging.FileHandler(filename, mode='w')
    file_handler.setLevel(logging.DEBUG)
    file_handler.setFormatter(get_default_formatter())

    logging.getLogger().addHandler(file_handler)


def get_default_formatter() -> logging.Formatter:
    """
    Return a default formatter for logging.

    The format is:
        Hours:Minutes:Second.Milliseconds - Location - [level]: message

    Returns:
        A logging.Formatter for the default format
    """
    return logging.Formatter(
        '%(asctime)s.%(msecs)03d - %(filename)s:%(lineno)d - '
        '[%(levelname)s]: %(message)s',
        datefmt='%H:%M:%S'
    )



"""
Version of the CLI file that includes PyCallGraph calls"""

from collections import OrderedDict
from contextlib import redirect_stdout

import numpy as np
import pandas as pd

from config import SETTINGS
from energy_hub import EHubModel

from pycallgraph import PyCallGraph
from pycallgraph import Config
from pycallgraph.output import GraphvizOutput


def print_section(section_name: str, solution_section: dict) -> None:
    """
    Print all the attributes with a heading.

    Args:
        section_name: The heading
        solution_section: The dictionary with all the attributes
    """
    half_heading = '=' * 10
    print(f"\n{half_heading} {section_name} {half_heading}")

    attributes = OrderedDict(sorted(solution_section.items()))
    for name, value in attributes.items():
        if isinstance(value, dict):
            value = pd.DataFrame.from_dict(value, orient='index')

            if list(value.columns) == [0]:
                value.columns = [name]

        print(f"\n{name}: \n{value}")


def pretty_print(results: dict) -> None:
    """Print the results in a prettier format.

    Args:
        results: The results dictionary to print
    """
    np.set_printoptions(linewidth=1000, suppress=True)

    version = results['version']
    solver = results['solver']

    print("Version: {}".format(version))

    print("Solver")
    print("time: {}".format(solver['time']))
    print("termination condition: {}".format(solver['termination_condition']))

    print("Solution")
    print_section('Objective', results['solution']['objective'])
    print_section('Sets', results['solution']['sets'])
    print_section('Parameters', results['solution']['parameters'])
    print_section('Variable Parameters', results['solution']['param_or_var'])
    print_section('Variables', results['solution']['variables'])


def main():
    config = Config(l='circo')
    graphviz = GraphvizOutput(output_file='filter_none.svg', font_size = 40, output_type = 'svg')

    with PyCallGraph(output=graphviz, config=config):
        """The main function for the CLI."""
        model = EHubModel(excel=SETTINGS["input_file"])

        results = model.solve()

        pretty_print(results)

        with open(SETTINGS["output_file"], 'w') as file:
            with redirect_stdout(file):
                pretty_print(results)


if __name__ == "__main__":
    main()



"""
A script for outputting the results of a PyEHub model.

Outputs can be stored to an excel spreadsheet or printed to the console.
"""

import collections
from collections import OrderedDict
from typing import Union
import pandas as pd
import matplotlib.pyplot as plt
from openpyxl import load_workbook

EXCEL_ENGINE = 'openpyxl'
EXCEL_FIRST_COLUMN_WIDTH = 25

def to_dataframes(frames: dict) -> OrderedDict:
    """
    Convert the values of a dictionary into dataframes if they can be converted.

    Args:
        frames: The dictionary to be converted

    Returns:
        An ordered dictionary with dataframes as values (if they can be
        converted).
    """
    for name, value in frames.items():
        if isinstance(value, dict):
            value = to_dataframe(name, value)

        frames[name] = value

    return OrderedDict(sorted(frames.items()))


def to_dataframe(name: str, value: dict) -> pd.DataFrame:
    """
    Convert a dictionary into a dataframe.

    Args:
        name: The name for the dataframe
        value: The dictionary to be converted

    Returns:
        A Pandas DataFrame.
    """
    value = sort_dict(value)
    value = pd.DataFrame.from_dict(value, orient='index')

    # To remove confusion on what the column '0' means
    if list(value.columns) == [0]:
        value.columns = [name]

    # Make wide matrices fit on the screen
    num_rows, num_columns = value.shape
    if num_columns > num_rows:
        value = value.T

    return value


def _create_sheet(excel_writer: pd.ExcelWriter, sheet_name: str):
    """
    Create an Excel sheet with the given name.

    Args:
        excel_writer: The ExcelWriter object
        sheet_name: The name of the new sheet

    Returns:
        The worksheet for the new sheet
    """
    # Create an empty dataframe just so that we can create a sheet
    pd.DataFrame().to_excel(excel_writer, sheet_name)

    return excel_writer.sheets[sheet_name]


def output_excel(solution: dict, file_name: str, time_steps: int = None, sheets:list = None):
    """ 
    Output the solution dictionary to an Excel file.

    It outputs the time series data in their own sheet with the rest being put
    into another sheet.

    Args:
        solution: A dictionary of the solution part of the response format.
            This contains the variables and parameters of the solved model.
        file_name: The name of the Excel file to write to
        time_steps: Optional. The number of time steps to classify a dataframe
            as holding time series data.
        sheets: A list of all the sheets to be contained in the excel file.
    """
    excel_writer = pd.ExcelWriter(file_name, engine=EXCEL_ENGINE)
    attributes = to_dataframes(solution)

    def _has_time_series_data(value):
        if isinstance(value, pd.DataFrame):
            num_rows, _ = value.shape

            return time_steps and num_rows >= time_steps
        return False

    time_series = {name: value for name, value in attributes.items()
                   if _has_time_series_data(value)}
    non_time_series = {name: value for name, value in attributes.items()
                       if name not in time_series}

    for name, value in time_series.items():
        value.to_excel(excel_writer, f'{name}')

    #: Set first_sheet to "Other" if sheets is not passed, otherwise set it to
    #: the first item in sheets.
    first_sheet = 'Other' if sheets is None else sheets[0]

    #: If sheets is passed, then pass all the sheets except the first as additional sheets.
    if sheets is not None:
        _dict_to_excel_sheet(excel_writer, first_sheet, non_time_series, padding=1, additional_sheets= sheets[1:])
    else:
        #: Otherwise dont sent any value for additional sheets, as none have been provided.
        _dict_to_excel_sheet(excel_writer, first_sheet, non_time_series, padding=1)

    excel_writer.save()


def _dict_to_excel_sheet(excel_writer, sheet_name, non_time_series, padding=0, additional_sheets: list=None):
    """ 
    Output a dictionary to an excel sheet.

    Args:
        excel_writer: The ExcelWriter object
        sheet_name: The name of the sheet to output to
        non_time_series: The dictionary of data
        padding: The padding between key, value pairs
        additional_sheets: a list containing names of other sheets
                           to potentially be added to the excel document.
    """
    hold_row = 1
    row = 1
    hold_sheet_name = sheet_name
    sheet = _create_sheet(excel_writer, sheet_name)
    hold_sheet = sheet

    for name, value in non_time_series.items():
        if additional_sheets is not None and name in additional_sheets:
            hold_row = row
            row = 1
            sheet = _create_sheet(excel_writer, name)
            sheet_name = name
        if isinstance(value, pd.DataFrame):
            num_rows, num_cols = value.shape

            # Compensate for empty Excel row of dataframes. No idea why. Gee thanks
            # Also handle case for when printing first value in sheet.
            startrow = max(row - 1, 1)
            value.to_excel(excel_writer, sheet_name, startrow=startrow)

            if num_cols > 1:
                # Output variable name if not in column header
                sheet.cell(row=row, column=1).value = name

            row += num_rows + 1  # Add row for column headers
        else:
            sheet.cell(row=row, column=1).value = name

            if not isinstance(value, collections.Iterable):
                # Simplify writing column values
                value = [value]

            for col, x in enumerate(value, start=2):
                sheet.cell(row=row, column=col).value = str(x)

            row += 1

        if sheet_name != hold_sheet_name:
            sheet_name = hold_sheet_name
            row = hold_row
            sheet = hold_sheet

        row += padding

    sheet.column_dimensions['A'].width = EXCEL_FIRST_COLUMN_WIDTH


def pretty_print(results: dict) -> None:
    """Print the results in a prettier format.

    Args:
        results: The results dictionary to print
    """
    version = results['version']
    solver = results['solver']

    print("Version: {}".format(version))

    print("Solver")
    print(f"\ttermination condition: {solver['termination_condition']}")
    if solver['termination_condition']!='Infeasible':
        print(f"\ttime: {solver['time']}")

        print("Solution")
        print_section('Stuff', results['solution'].copy())


def print_section(section_name: str, solution_section: dict) -> None:
    """
    Print all the attributes with a heading.

    Args:
        section_name: The heading
        solution_section: The dictionary with all the attributes
    """
    half_heading = '=' * 10
    print(f"\n{half_heading} {section_name} {half_heading}")

    attributes = to_dataframes(solution_section)
    for name, value in attributes.items():
        with pd.option_context('display.max_rows', None, 'display.max_columns', 3):
            print(f"\n{name}: \n{value}")


def sort_dict(mapping: Union[dict, OrderedDict]) -> OrderedDict:
    """
    Sorts a dictionary and all its sub-dicionaries as well.

    Examples:
        >>> sort_dict({1: 'a', 3: 'c', 2: 'b'})
        OrderedDict([(1, 'a'), (2, 'b'), (3, 'c')])
        >>> sort_dict({1: {3: 'c', 2: 'b', 1: 'a'}})
        OrderedDict([(1, OrderedDict([(1, 'a'), (2, 'b'), (3, 'c')]))])

    Args:
        mapping: The dictionary to be sorted

    Returns:
        The sorted dictionary as an OrderedDict
    """
    if not isinstance(mapping, dict):
        return mapping

    for key, value in mapping.items():
        mapping[key] = sort_dict(value)

    return OrderedDict(sorted(mapping.items()))


def print_capacities(results):
    """
    Prints the capacities of each tech and storage at the end

    :param results: the solved model

    """
    solution = results['solution']
    capacity_tech = solution['capacity_tech']
    capacity_storage = solution['capacity_storage']

    half_heading = '=' * 10
    print(f"\n{half_heading} {'Capacities'} {half_heading}")
    print(half_heading * 4)
    print(capacity_storage)
    print(half_heading * 4)
    print(capacity_tech)


def print_warning(results):
    """
    Prints an error if the model burns energy, i.e there is energy from storage and energy to storage
    at same time step.

    :param results: the solved model

    """
    solution = results['solution']
    storages = solution['storages']
    energy_from_storage = solution['energy_from_storage']
    energy_to_storage = solution['energy_to_storage']
    time = len(solution['time'])

    half_heading = '=' * 4

    for storage in storages:
        for t in range(time):
            energy_from = energy_from_storage[storage][t]
            energy_to = energy_to_storage[storage][t]

            if (energy_to != 0) and (energy_from != 0):
                print(f"{half_heading} {'Warning:'} {half_heading} \n{ 'Burning energy for storage: '} {storage} "
                      f"{'; For time step: '} {t} ")


def stream_info(results, output_file):
    """
    New output format with the information separated in different sheets for different streams.

    :param results: the solved model
    :param output_file: the output excel file
    """

    solution = results['solution']
    streams = solution['streams']
    load = solution['LOADS']
    energy_from_storage = solution['energy_from_storage']
    energy_to_storage = solution['energy_to_storage']
    energy_exported = solution['energy_exported']
    energy_imported = solution['energy_imported']
    demands = solution['demands']
    time = len(solution['time'])
    export_streams = solution['export_streams']
    import_streams = solution['import_streams']
    storages = solution['storages']
    storage_level = solution['storage_level']

    n = len(storages)


    with pd.ExcelWriter(output_file, engine=EXCEL_ENGINE) as writer:
        writer.book = load_workbook(output_file)

    keep = ['other', 'TIME_SERIES', 'energy_input', 'capacity_tech', 'capacity_storage']
    for sheet_name in writer.book.sheetnames:
        if sheet_name not in keep:
            del writer.book[sheet_name]

    for stream in streams:

        if stream in demands:
            df = pd.DataFrame({
                stream + " Load": [load[stream][t] for t in range(time)]})
            df.to_excel(writer, sheet_name=stream, startrow=time + 2)

        df1 = pd.DataFrame({
            "energy_from " + storage: [energy_from_storage[storage][t] for t in range(time)] for storage in storages})

        df2 = pd.DataFrame({
            "energy_to " + storage:  [energy_to_storage[storage][t] for t in range(time)] for storage in storages})

        df3 = pd.DataFrame({
            "level " + storage: [storage_level[storage][t] for t in range(time)] for storage in storages})

        df1.to_excel(writer, sheet_name=stream, )
        df2.to_excel(writer, sheet_name=stream, startcol=n + 2)
        df3.to_excel(writer, sheet_name=stream, startcol=2*(n + 2))

        if stream in import_streams:
            df4 = pd.DataFrame({
                stream + " imported": [energy_imported[stream][t] for t in range(time)]})
            df4.to_excel(writer, sheet_name=stream, startrow=time + 2, startcol=4)

        if stream in export_streams:
            df5 = pd.DataFrame({
                stream + " exported": [energy_exported[stream][t] for t in range(time)]})
            df5.to_excel(writer, sheet_name=stream, startrow=time + 2, startcol=7)

        writer.save()


def plot_storages(results: dict, **kwargs) -> None:
    """
    Plots various variables related to storages:
    storage state: The level of energy remaining in the storage.
    gross charge(charge from stream): Energy sent to storage from stream for charging.
    gross discharge(discharge to stream): Energy going into stream after discharging.
    net charge: Energy actually reaching storage(after loss of gross charge due to charging efficiency).
    net discharge: Energy actually leaving storage(after loss from this due to discharging efficiency, it becomes gross
                   discharge).
    decay loss(standing loss): Energy dissipating due to standing losses.
    Note: All of the above are plotted as default. You can change this by passing respective arguments.

    Args:
        results: dictionary; returned by solve() method.
        pl_state: boolean value; whether to plot 'storage state' or not.
        pl_gross_ch: boolean value; whether to plot 'gross charge' or not.
        pl_gross_dch: boolean value; whether to plot 'gross discharge' or not.
        pl_net_ch: boolean value; whether to plot 'net charge' or not.
        pl_net_dch: boolean value; whether to plot 'net discharge' or not.
        pl_decay: boolean value; whether to plot 'decay loss' or not.
        size: tuple; (width, height) of the plot [default is (10,5)].
        percentage: boolean; y-axis units in % or in kWh [default].
    """
    solution_section = results['solution'].copy()
    attributes = to_dataframes(solution_section)

    pl_state = kwargs.get('pl_state', True)
    pl_gross_ch = kwargs.get('pl_gross_ch', True)
    pl_gross_dch = kwargs.get('pl_gross_dch', True)
    pl_net_ch = kwargs.get('pl_net_ch', True)
    pl_net_dch = kwargs.get('pl_net_dch', True)
    pl_decay = kwargs.get('pl_decay', True)
    size = kwargs.get('size', (10, 5))
    if size[0] == 0 or size[1] == 0:
        raise ValueError("Please pass non-zero values in the 'size' tuple.")
    percentage = kwargs.get('percentage', False)

    for storage in attributes['storages']:
        fig, axes = plt.subplots(nrows=1, ncols=2, sharey=True, figsize=size)
        fig.text(0.524, 0.02, 'Timesteps (-)', ha='center', va='center')
        if percentage:
            fig.text(0.01, 0.5, 'Storage state [%]', ha='center', va='center', rotation='vertical')
        else:
            fig.text(0.01, 0.5, 'Storage state (kWh)', ha='center', va='center', rotation='vertical')
        fig.text(0.5, 0.97, storage, ha='center', va='center')
        ax = axes[0]
        ax1 = axes[1]


        ser_storage_state = attributes['storage_level'][storage]
        ser_gross_charge = attributes['energy_to_storage'][storage]
        ser_gross_discharge = -attributes['energy_from_storage'][storage]

        eff_charge = attributes['CHARGING_EFFICIENCY']['CHARGING_EFFICIENCY'][storage]
        eff_discharge = attributes['DISCHARGING_EFFICIENCY']['DISCHARGING_EFFICIENCY'][storage]
        decay = attributes['STORAGE_STANDING_LOSSES']['STORAGE_STANDING_LOSSES'][storage]
        capacity = attributes['capacity_storage']['capacity_storage'][storage]

        ser_net_discharge = 0
        ser_net_charge = ser_gross_charge.multiply(eff_charge)
        if eff_discharge != 0:
            ser_net_discharge = ser_gross_discharge.divide(eff_discharge)
        decay_loss = -ser_storage_state.multiply(decay)[:-1]

        d = {'storage state': ser_storage_state, 'charge from stream': ser_gross_charge, 'net charge': ser_net_charge,
             'discharge to stream': ser_gross_discharge, 'net discharge': ser_net_discharge, 'standing loss': decay_loss}
        df = pd.DataFrame(data=d)

        if percentage:
            if capacity != 0:
                df = (df.div(capacity)).mul(100)

        if pl_state:
            df['storage state'].plot(kind='bar',  title=' ', ax=ax, color='deepskyblue', legend=True)
        if pl_gross_ch:
            df['charge from stream'].plot(drawstyle='steps-post',  title=' ', color='lightgreen',
                                          linewidth=3, ax=ax1, legend = True)
        if pl_net_ch:
            df['net charge'].plot(drawstyle='steps-post',  title=' ', color='green',
                                  linestyle='--', linewidth=2, ax=ax1, legend=True)
        if pl_gross_dch:
            df['discharge to stream'].plot(drawstyle='steps-post',  title=' ', color='orange',
                                           linewidth=3, ax=ax1, legend=True)
        if pl_net_dch:
            df['net discharge'].plot(drawstyle='steps-post',  title=' ', color='red',
                                     linestyle='--', linewidth=2, ax=ax1, legend=True)
        if pl_decay:
            df['standing loss'].plot(kind='bar',  title=' ', color='saddlebrown',
                                     ax=ax, legend=True)
        ax1.set_xticks(df.index[:-1])
        ax.set_xticks(df.index)


        if pl_state or pl_gross_ch or pl_net_ch or pl_gross_dch \
                or pl_net_dch or pl_decay:
            plt.legend(loc='best')
            for ax in fig.axes:
                plt.sca(ax)
                plt.xticks(rotation=0)
            plt.show()


def plot_energy_balance(model, results: dict, **kwargs) -> None:
    """
    Visualization of energy balance for all the streams, i.e., Plots the energy interactions of all the streams with
    loads, converters, storages, imports and exports. Plots are dashed and marked with shape markers for visibility
    in case of overlap. Adjust the properties of the plot by passing arguments from below.

    Args:
        model: The object of EHubModel class (or any child class thereof); the energy hub model.
        results: dictionary; returned by solve() method.
        size: tuple; (width, height) of the plot [default is (9,5)].
        lw: float; linewidth of the plots [default is 2].
        dl: float; length of the dashes constituting the dashed plots [default is 3].
    """
    attributes = results['solution'].copy()
    streams_wo_sources = [x for x in model.streams if x not in model.sources]

    size = kwargs.get('size', (9, 5))
    if size[0] == 0 or size[1] == 0:
        raise ValueError("Please pass non-zero values in the 'size' tuple.")
    lw = kwargs.get('lw', 2)
    dl = kwargs.get('dl', 3)

    for stream in streams_wo_sources:
        fig, ax = plt.subplots(nrows=1, ncols=1, figsize=size)

        dict_data_pos = {}
        dict_data_neg = {}

        for t in model.time:
            curr_neg = curr_pos = 0
            if stream in model.demands:
                load = model.LOADS[stream][t]
                curr_neg += (-load)
                place_in_dict(t, model, dict_data_neg, 'LOAD', curr_neg)

            for storage in model._get_storages_from_stream(stream):
                q_out = float(attributes['energy_to_storage'][t][storage.name])
                curr_neg += (-q_out)
                q_in = float(attributes['energy_from_storage'][t][storage.name])
                curr_pos += q_in
                place_in_dict(t, model, dict_data_neg, 'To '+storage.name+' (S)', curr_neg)
                place_in_dict(t, model, dict_data_pos, 'From '+storage.name+' (S)', curr_pos)

            for tech in model.technologies:
                conversion_rate = float(attributes['CONVERSION_EFFICIENCY'][tech][stream])
                if conversion_rate < 0:
                    energy_input = float(attributes['energy_input'][t][tech])
                    curr_neg += (-energy_input)
                    place_in_dict(t, model, dict_data_neg, 'To '+tech+' (C)', curr_neg)

                if conversion_rate > 0:
                    energy_input = float(attributes['energy_input'][t][tech])
                    energy_output = energy_input * conversion_rate
                    curr_pos += energy_output
                    place_in_dict(t, model, dict_data_pos, 'From '+tech+' (C)', curr_pos)

            if stream in model.export_streams:
                energy_exported = attributes['energy_exported'][t][stream]
                curr_neg += (-energy_exported)
                place_in_dict(t, model, dict_data_neg, 'Exported', curr_neg)

            if stream in model.import_streams:
                energy_imported = attributes['energy_imported'][t][stream]
                curr_pos += energy_imported
                place_in_dict(t, model, dict_data_pos, 'Imported', curr_pos)

        # --------------------------------Plot Demands----------------------------------
        x = 1
        y = 1

        mk_c = ['1', '2', '3', '4']
        mk_s = ['*', 'P']

        mk_index_c = 0
        mk_index_s = 0

        for label, data in dict_data_neg.items():
            if label.endswith('LOAD'):
                ds = ()

                mk = 's'
                ms = 7

            if label.endswith('(C)'):
                ds = (dl, x)
                x += 1

                mk = mk_c[mk_index_c % 4]
                mk_index_c += 1
                ms = 12

            if label.endswith('(S)'):
                ds = (dl, y, 1, y)
                y += 1

                mk = mk_s[mk_index_s % 2]
                mk_index_s += 1
                ms = 7.5

            if label.endswith('Exported'):
                mk = '|'
                ds = (dl, 1, 1.5, 1, 1.5, 1, 1.5, 1, 1.5, 1, 1.5, 1)
                ms = 12

            ax.plot(data, label=label, linewidth=lw, dashes=ds, marker=mk, markersize=ms)

        handle_demand, label_demand = ax.get_legend_handles_labels()

        # --------------------------------Plot Supplies----------------------------------
        x = 1.2
        y = 1.2

        mk_c = ['1', '2', '3', '4']
        mk_s = ['+', 'x']

        mk_index_c = 0
        mk_index_s = 0

        for label, data in dict_data_pos.items():
            if label.endswith('(C)'):
                ds = (dl, x)
                x += 1

                mk = mk_c[mk_index_c % 4]
                mk_index_c += 1
                ms = 12

            if label.endswith('(S)'):
                ds = (dl, y, 1, y)
                y += 1

                mk = mk_s[mk_index_s % 2]
                mk_index_s += 1
                ms = 12

            if label.endswith('Imported'):
                ds = (dl, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1)
                mk = '|'
                ms = 12

            ax.plot(data, label=label, linewidth=lw, dashes=ds, marker=mk, markersize=ms)

        ax.set(title=stream, xticks=model.time, xlabel='Timesteps (-)', ylabel=stream + ' (kWh)')

        # To have 'Demands' and 'Supplies' titles, we need a blank handle
        blank_handle = [plt.plot([], marker="", ls="")[0]]

        handle_supply, label_supply = ax.get_legend_handles_labels()

        # Update handles and labels for 'Supplies' by removing 'Demands' handles and labels from them
        handle_supply = handle_supply[len(handle_demand):]
        label_supply = label_supply[len(label_demand):]

        handles = blank_handle + handle_demand + blank_handle*2 + handle_supply

        labels = ["$\\bf{Demands}$"] + label_demand + [" "] + ["$\\bf{Supplies}$"] + label_supply

        plt.legend(loc='best', bbox_to_anchor=(1, 1), labels=labels, handles=handles)
        plt.show()


def place_in_dict(t, model, dictionary: dict, key: str, value):
    """Appends an entry in the list corresponding to key if key exists in dictionary.
    Otherwise creates a single entry in the dictionary as {key: [value]}"""
    if t == model.time[0]:
        dictionary.update({key: [value]})
    else:
        dictionary[key].append(value)



"""
Provides functionality for handling a request format network_links.
"""
from typing import Union
class Network_links():
    """A wrapper for a request format Network_links."""

    def __init__(self, network_links_request: dict, capacity_link: dict) -> None:
        """Create a new wrapper for a network links.

        Args:
            network_links_request: the request format network links
            capacity_link:
        """
        self._links = network_links_request
        self._capacity = capacity_link


    @property
    def link_id(self) -> int:
        """return network link id"""
        return self._links['id']

    @property
    def start_id(self) -> int:
        """return start node id"""
        return self._links['start_id']

    @property
    def end_id(self) -> int:
        """return end node id"""
        return self._links['end_id']

    @property
    def length(self) -> float:
        """return the length of the link"""
        return self._links['length']

    @property
    def total_thermal_loss(self) -> float:
        """return the thermal loss of the connection"""
        return self._links['total_thermal_loss']

    @property
    def total_pressure_loss(self) -> float:
        """return the pressure loss of the connection"""
        return self._links['total_pressure_loss']

    @property
    def link_capacity(self) -> Union[float, str]:
        """return the capacity of the connection"""
        return self._links['capacity']

    @property
    def link_type(self) -> str:
        """return the type of the connection"""
        return self._links['type']

    @property
    def link_reactance(self) -> float:
        """return the reactance across the connection"""
        return self._links['reactance']
    # @property
    # def link_stream(self) -> str:
    #     """return stream associated with link"""
    #     return self._links['link_stream']

    @property
    def link_cost(self)->float:
        """return the cost of the connection"""
        return self._links['link_cost']



"""
A script used to solve a batch series of energy hubs in a combined PyEHub model.

This functionality is now also provided by the multiple hubs module.
"""

import argparse
import os
import re
import shutil
import openpyxl
from openpyxl import Workbook
from openpyxl.styles import Alignment
import multiple_hubs

""" File naming convention:

In order for an excel file to be read as input correctly, it must
follow the follwing naming convention:

[any valid file name]_[id to be assosiated with it].xlsx

note,
	carbon/cost is there to allow you to specify whether a test is trying to minimize
	either cost or carbon.

"""

def get_output_fields(filenames: list, directory: str = None):
	""" get_output_fields(list, str)

		Gets the required outputs from the excel files passed in the
		list of filenames.

		Args:
			filenames(list): A list of the excel files to find outputs for.

			directory(str): The optional dicretory that these filenames are located at.
					for use only if the filenames do not include their directory.

		Returns:
			A list with every output requested from every file (duplicates are deleted.)
	"""

	#: Create the list to hold the values in.
	master_field_list = []

	for filename in filenames:

		#: Open the excel file for this loop.
		if directory is not None:
			book = openpyxl.load_workbook(directory + filename)
		else:
			book = openpyxl.load_workbook(filename)

		#: Open the capacities sheet which is set to have all of
		#: the required outputs in it.
		sheet = book["Capacities"]

		#: Loop through the header (skipping not output fields)\
		#: and get all of the required outputs. These exist from
		#: (row 1, col 2) to (row 1, final col)
		fields = sheet.iter_rows(min_row=1, max_row=1, min_col=2)

		#: Convert the generator to a list, and get the only row.
		#: then loop through each cell in the row, and append the
		#: value in that cell to the master list.
		for cell in list(fields)[0]:
			master_field_list.append(cell.value)

	#: Append total_cost and total_cost to the output list so that these
	#: values are automatically included in the output.
	master_field_list.append("total_cost")
	master_field_list.append("total_carbon")
	#: Convert the list to a set to remove all duplicates,
	#: then convert the set back to a list, and return.
	return_list = []
	for item in master_field_list:
		if item is not None:
			return_list.append(item)

	return sorted(list(set(return_list)))


def output_to_excel(outputs: dict, sheet, output_dict: dict):
	""" output_to_excel(dict, openpyxl.sheet, dict)

		Takes the output dictionary produced by multiple hubs, and
		writes specific fields to the passed excel sheet based on
		the output dictionary passed.

		Args:
			outputs(dict): The dictionary produced by multiple_hubs.

			sheet(openpyxl.sheet): The excel sheet to be written to.

			output_dict(dict): A dictionary that links keys found in the
					   outputs dict, and the column that they should
					   appear in, within the sheet.

	"""

	#: Get all of the columns that have content in them, and use
	#: that to figure out where the next blank column is.
	cols = list(sheet.iter_cols(min_col=1, max_col=1))[0]
	first_empty_row = len(cols) + 1


	#: Merge the cells that have large words togther. These are the
	#: first 4 cells appearing in any output (Filename, funcion,
	#: fjob, and hub).
	#: File name header cell
	sheet.merge_cells("A{}:B{}".format(first_empty_row, first_empty_row))
	#: Function header cell
	sheet.merge_cells("C{}:D{}".format(first_empty_row, first_empty_row))
	#: fjob header cell
	sheet.merge_cells("E{}:F{}".format(first_empty_row, first_empty_row))
	#: Hub header cell
	sheet.merge_cells("G{}:H{}".format(first_empty_row, first_empty_row))

	#: for each key and value pair in the formatting dictionary passed.
	for key, value in output_dict.items():

		#: If outputs[key] exists (in other words if the current key from output_dict,
		#: represents an output that is in this set of outputs, then set it to the corrisponding
		#: column based off the value of this item.
		try:
			sheet['{}{}'.format(value, first_empty_row)] = outputs[key]

		#: Otherwise move on to the next loop, as this output was not outputted by this file.
		except:
			continue

def main(batch_run_location: str="excel_files/batch_input/", output_directory: str="batchrun_outputs/", epsilon_n: int = 0):

	#: If argparse got no -d switch
	if output_directory is None:

		#:Set output directory to its default setting.
		output_directory = "batchrun_outputs/"

	#: Otherwise, if the user forgot to / or \ terminate their directory, then
	#: do it for them.
	elif output_directory[-1] != "/" and output_directory[-1] != "\\":
		output_directory = output_directory + "/"

	#: If argparse got no -o switch
	if batch_run_location is None:

		#: Set batch_run_location to its default setting.
		batch_run_location = "excel_files/batch_input/"

	#: Otherwise if the user forgot to / or \ terminate their directory, then
	#: do it for them.
	elif batch_run_location[-1] != "/" and batch_run_location[-1] != "\\":
		batch_run_location = batch_run_location + "/"

	#: If argparse got no -n switch
	if epsilon_n is None:

		#: Set epsilon_n to iss default setting.
		epsilon_n = 0

	#: If the output directory is not currently a directory, make it.
	if not os.path.exists(output_directory):
		os.makedirs(output_directory)

	#: Get a list of all files in the run location that end in .xlsx (excel's file
	#: extension)
	excel_files = [file for file in os.listdir(batch_run_location) if file[-5:] == ".xlsx" and file[0] != '$']

	file_data = [] #: An intermediate container to hold some of the information
		       #: That will be gathered and held on the files.

	#: Get the id of every file found in the run location,
	#: then store a tuple with (filename, fileid) into file_data.
	for file in excel_files:

		#: Initial position while looking for the _.
		_position = 0

		#: For each character in the file name (backwards),
		#: if the character is an underscore then we know that
		#: this is the character before the id starts.
		for character in file[::-1]:

			#: Check if the character is an _, if so then
			#: we're done looking and should break from the
			#: inner loop.
			if character == "_":
				break

			#: Otherwise subtract 1 from position.
			_position -= 1

		#: Append a tuple defined by (filename, fileid).
		#: file[_position:-5] gets the id because _position
		#: is the location of the LAST _ in the filename (ie, the
		#: one that is required to preceed the id) and there are five
		#: characters at the end of each file ".xlsx". Thus the range
		#: between _position, and 5 is the id (no matter how many numbers.)
		file_data.append((file, file[_position:-5]))

	#: Sort file data based off of the file ids so that
	#: all pairs of files are beside eachother in the list.
	file_data = sorted(file_data, key=lambda x: x[1])

	file_sets = [] #: The final container to hold tuples of pairs of filenames.

	#: Append a tuple defined by (filename1, filename1's pair) to file_sets.
	#: for each pair in file_data.
	for index, file in enumerate(file_data[::2]):
		file_sets.append((file, file_data[(index * 2) + 1]))

	#: Get all the required outputs from the found excel files.
	all_fields = get_output_fields(excel_files, directory=batch_run_location)

	#: The first column to put data in is alwasy I
	first_letter = 'I'

	#: The number of fields automatically appeneded after the capacities.
	#: These fields are:
	#:	total_carbon
	#:	total_cost
	extra_fields = 2

	#: Calculate the letter(s) of the final column to put data in. - 1 is to
	#: account for an off by one error. The complexity in these functions is
	#: to account for the ability of the sheet to be over 26 columns long. In
	#: that case, excel starts using AA, then AB, etc so this must be accounted for.
	final_letter = str(first_letter)
	for _ in range(len(all_fields) - 1 - extra_fields):
		#: If the final letter is not a Z, then just increase it by a letter.
		if final_letter[-1] != 'Z':
			final_letter = final_letter[:-1] + chr(ord(final_letter[-1]) + 1)
		#: Otherwise remove the Z and replace it with AA to account for
		#: excel's column naming.
		else:
			final_letter = final_letter[:-1] + "AA"

	#: Create a list of the extended row's letter(s) representing their columns.
	extended = [final_letter]
	for _ in range(extra_fields):
		if extended[-1][-1] != 'Z':
			extended.append(extended[-1][:-1] + chr(ord(extended[-1][-1]) + 1))
		else:
			extended.append(extended[-1][:-1] + "AA")

	#: Create an excel workbook, and open a sheet.
	output_book = Workbook()
	sheet = output_book.active

	#: Merge cells for proper formatting of headers.
	#: Filename header
	sheet.merge_cells("A1:B2")
	#: Function header
	sheet.merge_cells("C1:D2")
	#: fjob header
	sheet.merge_cells("E1:F2")
	#: Hub header
	sheet.merge_cells("G1:H2")
	#: Capacity header: must span from first output to last output.
	sheet.merge_cells("{}1:{}1".format(first_letter, final_letter))
	#: Totals header must span from final_letter + 1 to final_letter + 2
	sheet.merge_cells("{}1:{}1".format(extended[1], extended[2]))

	#: Assign header values to these cells.
	#: Filename header
	sheet['A1'] = "File Name"
	#: Function header
	sheet['C1'] = "Function"
	#: fjob header
	sheet['E1'] = "fjob"
	#: Hub header
	sheet['G1'] = "Hub"
	#: Use first_letter here instead of 'I' to increase readability.
	sheet['{}1'.format(first_letter)] = "Capacity"
	#: Totals header
	sheet['{}1'.format(extended[1])] = "Totals"

	#: A dictionary describing which column data for each header should
	#: go in. Used as the output_dict parameter for output_to_excel().
	pos_dict = {
		     #: header: Data column,
			"Filename": "A",
			"Function": "C",
			"fjob": "E",
			"Hub": "G"
		}
	#: Create a string to parse through the data columnsm
	#: and to set values with.
	fletter = str(first_letter)

	#: For each field, append that field's name as a subheader under
	#: capacity.
	for field in all_fields:

		#: Append the current field name to the column
		#: stored currently in fletter.
		sheet['{}2'.format(fletter)] = field

		#: Build the dictionary used in output_to_excel().
		#: set the key value of field to the current letter
		#: stored by fletter.
		pos_dict[field] = fletter

		#: Increment fletter by one. If fletter's last (only) character
		#: is Z, then change it to AA
		if fletter[-1] != 'Z':
			fletter = fletter[:-1] + chr(ord(fletter[-1]) + 1)
		else:
			fletter = fletter[:-1] + "AA"

	#: Centre the cells all of the cells that we have used as headers.
	#: The Filename header
	sheet['A1'].alignment = Alignment(horizontal='center', vertical='center')
	#: The function header
	sheet['C1'].alignment = Alignment(horizontal='center', vertical='center')
	#: The fjob header
	sheet['E1'].alignment = Alignment(horizontal='center', vertical='center')
	#: The Hub header.
	sheet['G1'].alignment = Alignment(horizontal='center', vertical='center')
	#: Center the merged capacity cell.
	sheet['{}1'.format(first_letter)].alignment = Alignment(horizontal='center', vertical='center')
	#: Center the merged Totals cell.
	sheet['{}1'.format(extended[1])].alignment = Alignment(horizontal='center', vertical='center')

	#: Main execution loop. This handles the execution of each pair of files that was found.
	#: it also handles all of the outputting of the data found by the executions. Finally, it
	#: also handles looping over different carbon levels if epsilon_n is not 0.
	for fjob, file_set in enumerate(file_sets):

		#: Load the data from the first two excel files into the hub1 and hub2 files as required
		#: multiple_hubs.py
		shutil.copyfile(batch_run_location + file_set[0][0], 'docs/tutorials/network/hub1.xlsx')
		shutil.copyfile(batch_run_location + file_set[1][0], 'docs/tutorials/network/hub2.xlsx')

		#: Output current action as some actions may take a while and we want the user to know
		#: that we're not just frozen.
		print("Finding minimum cost for {} and {}".format(file_set[0][0][:-5], file_set[1][0][:-5]))

		#: Attempt to minimize the cost without worrying about the carbon output. Store the
		#: optimized data into cost_min.
		cost_min = multiple_hubs(False, output_directory + file_set[0][0][:_position - 1])

		#: Get which outputs are given by hub1 and hub2. NOTE these aren't the outputs,
		#: they are just the fields of important outputs to be logged in final_results.xlsx
		hub1_output_fields = get_output_fields([file_set[0][0]], directory=batch_run_location)
		hub2_output_fields = get_output_fields([file_set[1][0]], directory=batch_run_location)

		#: Create the base dictionary to output data with. This dictionary will
		#: be used as the outputs variable for output_to_excel().
		#: This base dictionary creates the information that isnt returned by
		#: multiple_hubs; this is mostly just readability data (ie used to distinguish
		#: between different runs.)
		hub1_outputs = {
				"Filename": file_set[0][0][:-5],
				"Function": "Minimize Cost",
				"fjob": (2 * fjob) + 1,
				"Hub": 1
			}

		#: Create the base dictionary to output data with. This dictionary will
		#: be used as the outputs variable for output_to_excel().
		#: This base dictionary creates the information that isnt returned by
		#: multiple_hubs; this is mostly just readability data (ie used to distinguish
		#: between different runs.)
		hub2_outputs = {
				"Filename": file_set[1][0][:-5],
				"Function": "Minimize Cost",
				"fjob": (2 * fjob) + 1,
				"Hub": 2
			}

		#: For each field that we need to store for hub 1,
		#: add an entry to hub1_outputs where the key is
		#: field, and the value is the value for that field
		#: found by running multiple_hubs()
		for field in hub1_output_fields:
			hub1_outputs[field] = cost_min[0][field]

		#: For each field that we need to store for hub 2,
		#: add an entry to hub1_outputs where the key is
		#: field, and the value is the value for that field
		#: found by running multiple_hubs()
		for field in hub2_output_fields:
			hub2_outputs[field] = cost_min[1][field]

		#: Output all of the data found to the working excel file.
		output_to_excel(outputs=hub1_outputs, sheet=sheet, output_dict=pos_dict)
		output_to_excel(outputs=hub2_outputs, sheet=sheet, output_dict=pos_dict)

		#: Find the average carbon used on a run that minimizes cost (and theoretically
		#: maximizes carbon.)
		avg_carbon = (cost_min[0]["total_carbon"] + cost_min[1]["total_carbon"]) / 2

		#: Find the value of the carbon that should be used as a maximum when
		#: multiplied by the current step.
		carbon_per_step = 0 if epsilon_n == 0 else avg_carbon / epsilon_n

		#: Handles the epsilon function.
		#: If epsilon_n is 0, then this will only run once in an attempt to
		#: minimize cost at certain (n * carbon_per_step) carbon maximums.
		#Otherwise, we will run espilon_n times + 1 for the minimize carbon run.
		for n in range(epsilon_n + 1):

			#: Print which files and what n value we're currently at (also print
			#: the maximum carbon for this run.)
			print("Running {} and {}  at n = {}, max carbon = {}"\
				.format(file_set[0][0][:-5], file_set[1][0][:-5], n, carbon_per_step * n))

			#: Attempt to minimize the cost, with a maximum carbon value of carbon_per_step * n.
			#: Note,
			#:	If this is the first/only loop then carbon will be maxed at zero, while attempting
			#:	to optimize for cost. This will simulate a no-carbon building, and will theoretically
			#:	be the maximum cost. On all other runs, the carbon will be maxed at carbon_per_step * n.
			carbon_min = multiple_hubs(True, output_directory + "n-{}".format(n) + file_set[0][0][:_position - 1],
						   carbon_per_step * n)

			#: Create the base dictionary to output data with. This dictionary will
			#: be used as the outputs variable for output_to_excel().
			#: This base dictionary creates the information that isnt returned by
			#: multiple_hubs; this is mostly just readability data (ie used to distinguish
			#: between different runs.)
			hub1_outputs = {
					"Filename": file_set[0][0][:-5],
					"Function": "Minimize Carbon; n = {}".format(n),
					"fjob": (fjob + 1) * 2,
					"Hub": 1
				}

			#: Create the base dictionary to output data with. This dictionary will
			#: be used as the outputs variable for output_to_excel().
			#: This base dictionary creates the information that isnt returned by
			#: multiple_hubs; this is mostly just readability data (ie used to distinguish
			#: between different runs.)
			hub2_outputs = {
					"Filename": file_set[1][0][:-5],
					"Function": "Minimize Carbon; n = {}".format(n),
					"fjob": (fjob + 1) * 2,
					"Hub": 2
				}

			#: For each field that we need to store for hub 1,
			#: add an entry to hub1_outputs where the key is
			#: field, and the value is the value for that field
			#: found by running multiple_hubs()
			for field in hub1_output_fields:
				hub1_outputs[field] = carbon_min[0][field]

			#: For each field that we need to store for hub 2,
			#: add an entry to hub1_outputs where the key is
			#: field, and the value is the value for that field
			#: found by running multiple_hubs()
			for field in hub2_output_fields:
				hub2_outputs[field] = carbon_min[1][field]

			#: Output all of the data found to the working excel file.
			output_to_excel(outputs=hub1_outputs, sheet=sheet, output_dict=pos_dict)
			output_to_excel(outputs=hub2_outputs, sheet=sheet, output_dict=pos_dict)

	#: Save the excel spread sheet that was used to store all of the runs's important outputs.
	output_book.save(output_directory + "final_results.xlsx")

if __name__ == "__main__":

	#: Create the argument parser to parse the passed arguments.
	parser = argparse.ArgumentParser(description="A commandline utility to batch run tests.")

	#: Add the -d argument.
	parser.add_argument('-d', '--directory_name', type=str, help='The directory containing the excel file to be run.\n'\
									'Default is excel_files/batch_input/')
	#: Add the -o argument.
	parser.add_argument('-o', '--output_name', type=str, help='The directory to store all outputted files in..\n'\
									'Default is batchrun_outputs/')
	#: Add the -n argument.
	parser.add_argument('-n', '--epsilon_n', type=int, help='The number of steps to break into for the epsilon constraint.')

	#: Parse the passed arguments.
	args = parser.parse_args()

	#: Pass the arguments that were passed.
	main(args.directory_name, args.output_name, args.epsilon_n)



"""
Provides functionality for handling the streams in the request format.
"""


class Stream:
    """A wrapper for a request format stream."""

    def __init__(self, stream_request: dict, request: dict) -> None:
        """
        Create a new wrapper.

        Args:
            stream_request: The request format stream
            request: The request format
        """
        self._stream = stream_request
        self._request = request

    @property
    def co2(self) -> float:
        """The C02 factor of the stream."""
        return self._stream['co2']

    @property
    def co2_credit(self) -> float:
        """The C02 factor of the stream."""
        return self._stream['co2_credit']

    @property
    def importable(self) -> bool:
        """If the stream is importable."""
        return self._stream['importable']

    @property
    def exportable(self) -> bool:
        """If the stream is exportable."""
        return self._stream['exportable']

    @property
    def timeseries(self) -> str:
        """The availability of the stream."""
        return self._stream['timeseries']

    @property
    def price(self) -> float:
        """The price of the stream."""
        return self._stream['price']

    @property
    def export_price(self) -> float:
        """The export price of the stream."""
        return self._stream['export_price']

    @property
    def is_output(self) -> bool:
        """Is this an output stream?"""
        for tech in self._request['converters']:
            if self.name in tech['outputs']:
                return True

        return False

    @property
    def name(self) -> str:
        """The name of the stream."""
        return self._stream['name']



"""
Provides functionality for handling a request format's time series.
"""
from typing import Dict, List


class TimeSeries:
    """A wrapper for a request format time series."""

    def __init__(self, time_series_request: dict) -> None:
        """
        Create a new wrapper.

        Args:
            time_series_request: The request format time series
        """
        self._series = time_series_request

    @property
    def stream(self) -> str:
        """Return the name of the stream."""
        return self._series['stream']

    @property
    def is_source(self) -> bool:
        """Is this time series a source?"""
        return self._series['type'] == 'Source'

    @property
    def is_demand(self) -> bool:
        """Is this time series a demand?"""
        return self._series['type'] == 'Demand'

    @property
    def is_price(self) -> bool:
        """Is this time series a price"""
        return self._series['type'] == 'Price'

    @property
    def name(self) -> str:
        """The name of the time series."""
        return self._series['id']

    @property
    def is_solar(self) -> bool:
        """Is this time series for solar data?"""
        return self.is_source and (self._series['id'] == 'Solar'
                                   or self._series['id'] == 'Irradiation')

    @property
    def data(self) -> Dict[int, float]:
        """A dictionary from time to the values of the series."""
        return {
            column: value
            for column, value in enumerate(self._series['data'])
        }



"""
A script for testing the batchrun script's code.
"""

import subprocess
import openpyxl as op
import numpy as np

LETTERS = "ABCDEFGHIJKLMNOPQRSTUVWXYZ"

# Step 1: set the start and stop point of changing parameter and number of steps(+1).
# Step 2: Set the sheet name and cell number of parameter (Sheet_name, cell_to_set).
# Step 3: In (in_file), write the directory of the input excel.
# Step 4: In (model_out_dir), put the output directory which want to save in.
# Step 5: model_inputs is the directory that the temporary file saves in. It doesn't need to change it.
# Step 6: The summary output file which contains total cost and carbon will save in 'final_output_file.
# You can change its name here.
# Step 7: in model_out_file, you can change the name of output file per each run.


def main():
    # This is the price ranges for diesel Fuel in $/kWh
    start_D = 0.1
    stop_D = 0.5
    steps_D = 9

    start_PV = 50
    stop_PV = 250
    steps_PV = 6

    in_file = 'Individual_Hubs/PV_Diesel/RUN2/Input_Diesel_PV_Tanklimit.xlsx'
    sheet_name = 'Streams'
    cell_to_set = 'D5'

    sheet_name_PV = 'Converters'
    cell_to_set_PV='B4'

    model_out_dir = 'Individual_Hubs/PV_Diesel/RUN2'
    model_inputs = 'Individual_Hubs/PV_Diesel/RUN2/temp_inputs.xls'
    sheet_to_read = 'Other'
    cells_to_read = ['B50', 'B52','B54','B60','B110','B168', 'B170']
    final_output_file = 'Individual_Hubs/PV_Diesel/RUN2/output_Diesel_Tanklimit.xlsx'
    result_wb = op.Workbook()
    result_sheet = result_wb['Sheet']
    output_cells = ['Input value'] + cells_to_read
    for column, cell in zip(LETTERS, output_cells):
        result_sheet[f'{column}1'].value = cell

    for roww,valuee in enumerate(np.linspace(start_PV, stop_PV, steps_PV), start=2):
        wb = op.load_workbook(in_file)
        wb[sheet_name_PV][cell_to_set_PV].value = valuee

        for row,value in enumerate(np.linspace(start_D, stop_D, steps_D), start=2):
            wb = op.load_workbook(in_file)
            wb[sheet_name][cell_to_set].value = value
            wb.save(model_inputs)
            model_out_file = fr'{model_out_dir}/PV{valuee}_D{value}_TL.xlsx'
            #the input file for run.py is specified in config.yaml
            subprocess.run(['python', 'run.py', '--output', model_out_file])


            wb = op.load_workbook(model_out_file)
            result_sheet[f'A{row}'].value = value
            for col,cell_to_read in zip(LETTERS[1:], cells_to_read):
                print(f'writing {wb[sheet_to_read][cell_to_read].value} to {col}{row} for {value}')
                result_sheet[f'{col}{row}'].value = wb[sheet_to_read][cell_to_read].value
        result_wb.save(final_output_file)




if __name__ == '__main__':
    main()



"""
Solving energy hub model for n number of hubs with any network of connections wanted between the hubs.

To run for multiple hubs:
    $ python multiple_hubs.py -n NUMBER OF HUBS

If n (NUMBER OF HUBS) is not inputted the code will not run.

Naming input excel files:
    file names should start with "hub" followed by increasing numbers starting from 0.
    The files should be in the 'network' folder

To set the connections between the hubs ( the links):
   All the links between the hubs should be set in a separate excel file in the 'network' folder.
   It should be called "network.xlsx"

   The default set up is one directional connections.
   To set bidirectional connections, the connections should be defined per each direction, i.e:

        one direction:  link 0:  start_id: 0   end_id: 1


        bidirectional:  link 0:  start_id: 0   end_id: 1,
                        link 1:  start_id: 1   end_id: 0

    The link ids should start from 0 and increase by 1.
    The node ids (start_id and end_id) correspond to hubs numbering in the names of the hubs excel files.

Note:
Do not name constraints specific names in the EHubModel class -> will not be able to construct constraints in multiple_hubs
"""
import argparse
import os
import numpy as np

from energy_hub.param_var import ConstantOrVar
from energy_hub import EHubModel
from energy_hub.utils import constraint
from energy_hub.input_data import InputData
from outputter import print_section, output_excel
import network_to_request_format
import pylp
from pylp import RealVariable, BinaryVariable, IntegerVariable

import pdb


DOMAIN_TO_VARIABLE = {
    'Continuous': RealVariable,
    'Integer': IntegerVariable,
    'Binary': BinaryVariable,
}




class NetworkModel(EHubModel):
    """
    A subclass that allows connections between hubs.
    """

    def __init__(self, *, excel=None, request=None, name=None, network=None, network_request=None, hub_id=None, max_carbon=None):
        super().__init__(excel=excel, request=request, max_carbon=max_carbon)
        self.name = name
        self.hub_id = hub_id
        if network:
                network_request = network_to_request_format.convert(network)

        if network_request:
            self._net_data = InputData(network_request)
        
        if self._net_data:
            self.links = self._net_data.links_ids
            self.link_capacities = ConstantOrVar(self.links, model=self, values=self._net_data.link_capacity)
            self.network_cost = RealVariable()
            self.is_link_installed = {link: BinaryVariable()
                                    for link in self.links}

            self.LINK_LENGTH = self._net_data.link_length
            self.FIXED_NETWORK_INVESTMENT_COST = self._net_data.fixed_network_investment_cost
            self.LINK_PROPORTIONAL_COST = self._net_data.link_proportional_cost
        
        else:
            raise RuntimeError("Can't create a network with no data.")

    def _add_variables(self):
        super()._add_variables()
        self._add_link_capacity_variables()
        for i in self._net_data.capacities:
            self._data.capacities.append(i)

    @constraint('links')
    def link_is_installed(self, link):
        """
        Set binary to 1 if capacity of link is > 0.
        Args:
            link: A link
        """
        capacity = self.link_capacities[link]
        rhs = self.BIG_M * self.is_link_installed[link]
        return capacity <= rhs

    @constraint('links')
    def link_is_installed_2(self, link):
        """
        Set binary to 1 if capacity of link is > 0.
        Args:
            link: A link
        """
        installed = self.is_link_installed[link]
        lhs = self.BIG_M * self.link_capacities[link]
        return lhs >= installed

    @constraint()
    def calc_network_investment_cost(self):
        """
        Calculating investment cost for the links the hub has.
        Cost split between 2 hubs that are connected with the same link

        """
        cost = 0
        for i, link in enumerate(self.links):
            if (self._net_data.link_start[i] == self.hub_id) or (self._net_data.link_end[i] == self.hub_id):
                cost += (self.FIXED_NETWORK_INVESTMENT_COST * self.LINK_LENGTH[link] * self.is_link_installed[link]
                         + self.link_capacities[link] * self.LINK_LENGTH[link] * self.LINK_PROPORTIONAL_COST)

        cost = cost/2
        return self.network_cost == cost

    @constraint()
    def calc_total_cost(self):
        """
        Modifying total cost constraint. Adding network cost to the total cost.
        """
        parent_constraint = super().calc_total_cost()
        old_cost = parent_constraint.rhs

        return self.total_cost == old_cost + self.network_cost

    def _add_link_capacity_variables(self):
        for capacity in self._net_data.capacities:
            domain = capacity.domain
            name = capacity.name

            try:
                variable = DOMAIN_TO_VARIABLE[domain]()
            except KeyError:
                raise ValueError(
                    f'Cannot create variable of type: {domain}. Can only be: '
                    f'{list(DOMAIN_TO_VARIABLE.keys())}'
                )

            setattr(self, name, variable)

# NETWORK Link specific constraints:

@constraint()
def network_constraint(hub, link_end, link_start):
    """
    Yields the constraints that allow a network connection between two hubs.

    Args:
        hub: The hub
        link_end: all the links that the hub ends at
        link_start: all the links that the hub start at
    Yields:
       A network energy balanced constraints for each hub
    """

    for t in hub.time:
        link_starting = []
        link_ending = []

        for i in range(len(link_start)):
            link_starting.append(link_start[i][t])

        for i in range(len(link_end)):
            link_ending.append(link_end[i][t])

        yield ((hub.energy_exported[t]['Net_export'] - hub.energy_imported[t]['Net_import'] + sum(link_ending)
                - sum(link_starting)) == 0)

@constraint()
def power_balance_constraint(hub_index, hubs, Bmatrix, angles):

    for t in hubs[0].time:
        # P_{ij}=sum_{i} B_{ij}*theta_j ???
        
        line_flows_from_hub = sum([Bmatrix[hub_index][other_hub_index] * angles[other_hub_index][t] for other_hub_index in range(len(hubs))])
        yield sum([-line_flows_from_hub, hubs[hub_index].energy_exported[t]['Net_export'] - hubs[hub_index].energy_imported[t]['Net_import']]) == 0
        # yield sum([-line_flows_from_hub, hubs[hub_index].energy_exported[t]['Net_export'], -hubs[hub_index].energy_imported[t]['Net_import']]) == 0
        

@constraint()
def swing_hub_constraint(hub_index, hubs, angles):
    for t in hubs[hub_index].time:
        yield angles[hub_index][t] == 0

@constraint()
def link_capacity_constraint(link, hub, i):
    """
    Constraint for the flow in the links.
    """
    for flow in link:
        yield flow >= 0
        yield flow <= hub.link_capacities[i]

@constraint()
def linear_power_flow_constraint(power, angle_from, angle_to, time, reactance):
    """
    Constraint for linear powerflows
    """
    
    #FIXME: need to validate this constraint
    for t in time:
        yield power[t] == 1/reactance * (angle_from[t] - angle_to[t])
    

def multiple_hubs(minimize_carbon=False, output_filename=None, input_files=None, network_excel=None, network_request=None, max_carbon=None, n=0, solver='glpk'):
    """
    Core function for solving of multiple PyEHub models.
    """

    if network_excel:
        network_excel_file = network_excel
        network_request = network_to_request_format.convert(network_excel_file)

        if network_request:
            _net_data = InputData(network_request)
        else:
            # The default excel file for network, in the network directory
            # network_excel_file = 'Input_files/network_input/network.xlsx'
            # network_request = network_to_request_format.convert(network_excel_file)
            # _net_data = InputData(network_request)
            raise RuntimeError("Can't create a network with no network data.")

    # Create all the hubs
    hubs = []
    for i in range(0, n):
        # Default files in the network directory
        if input_files is None:
            file_name = f'Input_files/network_input/hub{i + 1}.xlsx'
        # Additional options for test for multiple hubs. To input a different directory and excel files
        else:
            file_name = f'{input_files}{i+1}.xlsx'
        excel_file = file_name

        hub = NetworkModel(excel=excel_file, name=f'hub{i+1}', max_carbon=max_carbon,
                                            network=network_excel_file, network_request=network_request, hub_id=i)
        hubs.append(hub)

    # Creating a list of all the constraints from all the hubs
    constraints = []
    for hub in hubs:
        hub.recompile()
        for constr in hub.constraints:
            constraints.append(constr)

    # The connections between the hubs, from the excel network input file.
    connections = []
    for i in _net_data.links_ids:
        start = _net_data.link_start[i]
        end = _net_data.link_end[i]
        link_type = _net_data.link_type[i]
        reactance = _net_data.link_reactance[i]

        connections.insert(i, (start, end, link_type, reactance))
    
    print(connections)
    
    links = []

    # The energy flow in each link per each time step
    energy_flow = {
        t: {out: RealVariable() for out in range(len(connections))}
        for t in hubs[0].time
    }
    

    # TODO: The voltage angle for each node on the elec powerflows 
    # Need to ensure that duplicates of the same node are not generated
    # Could be done when we append the link_stqarts and ends to each hub
    # The hubs are the nodes, just create a RealVariable() with a time for each hub
    # Every hub gets a voltage angle, but then they are only added to the model if they have a link of elec type
    
    num_hubs = len(hubs)
    default_reactance = 0.05

    voltage_angle = {
        t: {node: RealVariable() for node in range(num_hubs)}
        for t in hubs[0].time
    }
    
    angles = []

    # Constructing the list of voltage angles for the different hubs(nodes)
    for i in range(num_hubs):
        hub = hubs[0]
        angle = []
        for t in hub.time:
            angle.append(voltage_angle[t][i])
        angles.insert(i, angle)


    # Constructing list of energy flows for the links
    for i in range(len(connections)):
        hub = hubs[0]
        flow = []
        for t in hub.time:
            flow.append(energy_flow[t][i])

        links.insert(i, flow)
    
    # Creating admittance Matrix for each hub to handle power (modified from SILVER_VER_17's minpower's powersystem)
    Bmatrix = np.zeros((num_hubs, num_hubs))
    for i in range(len(connections)):
        if connections[i][2].lower() == 'power':
            try:
                # To grab the electrical reactance of the link if it exists
                link_reactance = float(connections[i][3])
            except ValueError:
                # There isn't a proper value (probably left blank) so use some sort of default value
                link_reactance = default_reactance
            hub_from_id = connections[i][0]
            hub_to_id = connections[i][1]
            Bmatrix[hub_from_id, hub_to_id] += -1 / link_reactance
            Bmatrix[hub_to_id, hub_from_id] += -1 / link_reactance
        for j in range(0, num_hubs):
            Bmatrix[j, j] = -1 * sum(Bmatrix[j, :])

    # Linear Power flow
    for i in range(len(connections)):
        #ensuring its a power transmission line
        
        if connections[i][2].lower() == 'power':
            try:
                # To grab the electrical reactance of the link
                link_reactance = float(connections[i][3])
            except ValueError:
                # There isn't a proper value (probably left blank) so use some sort of default value 
                link_reactance = default_reactance
            angle_from = angles[connections[i][0]]
            angle_to = angles[connections[i][1]]
            power = links[i]
            time = hubs[0].time
            for c in linear_power_flow_constraint(power, angle_from, angle_to, time, link_reactance):
                
                constraints.append(c)
                # pass

    # Creating a constraint that sets the bounds of the flow in the links:
    # link[i] >= 0. To ensure one directional connection
    # link[i] <= link_capacity.
    for hub in hubs:
        for i, link in enumerate(links):
            for c in link_capacity_constraint(link, hub, i):
                constraints.append(c)

    LINK_THERMAL_LOSS = _net_data.link_thermal_loss

    # Swing Bus
    if len(hubs) > 1:
        for c in swing_hub_constraint(0,hubs,angles):
            constraints.append(c)

    # Create a network connection
    for k, hub in enumerate(hubs):
        # New list for all the links that the hub starts at
        link_start = []
        # New list for all the links that the hub ends at
        link_end = []

        power_line_flag = 0

        for i in range(len(connections)):
            # Searching for the links that the hub starts at but are not power flow 
            if connections[i][2].lower == 'power':
                power_line_flag = 1
            else:
                if connections[i][0] == k:
                    link_start.append(links[i])
                # Searching for the links that the hubs ends at but are not power flow
                if connections[i][1] == k:
                    # Multiplying by thermal loss to account for the thermal loss in the link
                    links[i] = np.array(links[i])
                    links[i] = links[i] * LINK_THERMAL_LOSS[i]
                    link_end.append(links[i])

        # Creating the network constraints for the hubs:
        # Energy balancing for none powerflow links
        if power_line_flag == 0:
            for c in network_constraint(hub, link_end, link_start):
                constraints.append(c)
        if power_line_flag == 1:
            # Power balancing for transmission links
            for c in power_balance_constraint(k, hubs, Bmatrix, angles):
                constraints.append(c)


    # The objective function of the model is the summation of all the hub's
    # objective function
    objective = hubs[0].objective
    for hub in hubs[1:]:
        objective += hub.objective

    # Now solve this model.
    status = pylp.solve(objective=objective, constraints=constraints, minimize=True, solver=solver)

    #: Define the different sheets that we output
    sheets = [
        "Other",
   	    "LOADS",
   	    "SOLAR_EM",
   	    "energy_exported",
   	    "energy_from_storage",
   	    "energy_imported",
   	    "energy_to_storage",
   	    "is_on",
   	    "storage_level",
        "capacity_storage",
        "capacity_tech",
        ]

    #TODO: clean up the use of this input variable to just checking the objective of the hubs instead.
    #: Used to output files with names according to their job.
    if minimize_carbon:
        ext = "_minimized_carbon"
    else:
        ext = "_minimized_cost"

    #: A list used to return all dictionaries gathered from this function.
    hub_dict_list = []

    # Print out the hub's name and all stuff related to that model.
    tcar = 0
    tcost = 0
    tnetwork = 0
    for i, hub in enumerate(hubs):
        hub_dict_list.append(hub.solution_dict)
        if output_filename is None:
            output_excel(hub.solution_dict, f'{hub.name}{ext}.xlsx', sheets=sheets)
            print_section(f'{hub.name}{ext}', hub.solution_dict)
        else:
            output_excel(hub.solution_dict, f'{output_filename}_{i}{ext}.xlsx', sheets=sheets)
            print_section(f'{hub.name}_{i}{ext}', hub.solution_dict)
        tcar += hub.solution_dict["total_carbon"]
        tcost += hub.solution_dict["total_cost"]
        tnetwork += hub.solution_dict["network_cost"]

    for i, hub in enumerate(hubs):
        half_heading = '=' * 4
        for t in hub.time:
            net_imp = hub.solution_dict['energy_imported'][t]['Net_import']
            net_exp = hub.solution_dict['energy_exported'][t]['Net_export']

            if (net_imp != 0) and (net_exp != 0):
                print(f"{half_heading} {'Warning:'} {half_heading} \n{'Export & import at the same time for hub: '} {i}"
                      f"{'; For time step: '} {t} ")

    print("\nAbsolute total carbon:\t" + str(tcar))
    print("Absolute total cost:\t" + str(tcost), end='\n\n')
    for hub in hubs:
        print(hub.name + " network cost: " + str(hub.solution_dict['network_cost']))
    print("\nAbsolute total network cost:\t" + str(tnetwork))
    return hub_dict_list


if __name__ == '__main__':

    parser = argparse.ArgumentParser(description="Command line utility to run tests.")
    parser.add_argument('-c', '--carbon', action='store_true', help='Optimize based on the total carbon level.')
    parser.add_argument('-o', '--output_filename', type=str, help='The name that the output should be stored under.')
    parser.add_argument('-v', '--carbon_value', type=float, help='The value that the carbon should try to be.')
    parser.add_argument('-n', '--number_of_hubs', type=int, help='The number of hubs to be used')
    parser.add_argument('-i', '--input_files', type=str, help='The name of the directory and the beginning of the'
                                                              'the input files ')
    parser.add_argument('--network', type=str, help='The name that the network excel file')

    args = parser.parse_args()

    multiple_hubs(args.carbon, args.output_filename, args.input_files, args.network,
                  args.carbon_value, args.number_of_hubs)



"""
The package of everything EHub related.
"""
import os
import sys
import inspect

# import pyehub
# name = "pyehub"

CURRENT_DIR = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))
sys.path.append(CURRENT_DIR)



"""
Provides functionality for handling the response format.

Examples:

    If you want to validate a dictionary against the response format::

    >>> from data_formats import response_format
    >>> example = {}
    >>> response_format.validate(example)
    Traceback (most recent call last):
        ...
    data_formats.response_format.ResponseValidationError
"""
from typing import Dict, Any, Union

import jsonschema

from energy_hub.param_var import ConstantOrVar
from pylp import Variable, Status

SCHEMA = {
    'type': 'object',
    'properties': {
        'version': {
            'description': 'The SemVer version number for the format',
            'type': 'string',
            'enum': ['0.1.0'],
        },
        'solution': {
            'type': 'object',
        },
        'solver': {
            'type': 'object',
            'properties': {
                'termination_condition': {
                    'type': 'string',
                },
                'time': {
                    'type': 'number',
                    'minimum': 0,
                },
            },
            'required': ['termination_condition', 'time'],
            'additionalProperties': False,
        },
    },
    'required': ['version', 'solution', 'solver'],
    'additionalProperties': False,
}


class ResponseValidationError(Exception):
    """The instance failed to validate against the SCHEMA."""


def _is_valid_json_type(json: Any) -> bool:
    if json is None:
        return True

    return any(isinstance(json, t) for t in [str, float, int, dict, list])


def _is_valid(mapping: Union[dict, list, str, float, int]) -> bool:
    if isinstance(mapping, list):
        return all(_is_valid(x) for x in mapping)
    elif isinstance(mapping, dict):
        return all(_is_valid(key) and _is_valid(value)
                   for key, value in mapping.items())

    return _is_valid_json_type(mapping)


def validate(instance: dict) -> None:
    """
    Validate the instance against the schema.

    Args:
        instance: The potential instance of the schema

    Raises:
        ValidationError: the instance does not match the schema
    """
    if not _is_valid(instance):
        raise ResponseValidationError

    try:
        jsonschema.validate(instance, SCHEMA)
    except jsonschema.ValidationError as exc:
        raise ResponseValidationError from exc


def _get_value(value: Any) -> Any:
    if isinstance(value, Variable):
        return value.evaluate()
    elif isinstance(value, range):
        return list(value)
    elif isinstance(value, ConstantOrVar):
        return _get_value(value.values)
    elif isinstance(value, dict):
        return {key: _get_value(value) for key, value in value.items()}

    return value


def _get_stuff(model: Dict) -> Dict[str, Any]:
    return {name: _get_value(value) for name, value in model.items()}


def create_response(status: Status, model: Dict) -> Dict[str, Any]:
    """
    Create a new response format dictionary.

    Args:
        status: The status of the problem
        model: A dictionary of the variables and their values that were used in
            the model
    """
    result = {
        'version': '0.1.0',
        'solver': {
            'termination_condition': status.status,
            'time': status.time,
        },
        'solution': _get_stuff(model),
    }

    validate(result)

    return result



"""
Provides functionality for handling the request format.

Examples:

    If you want to validate a dictionary against the request format::

    >>> from data_formats import request_format
    >>> example = {}
    >>> request_format.validate(example)
    Traceback (most recent call last):
        ...
    data_formats.request_format.request_format.RequestValidationError
"""
import jsonschema

SCHEMA = {
    'title': 'EHub Model Request Format',
    'description': 'A JSON Schema used to validate an incoming request to '
                   'solve a model',
    'type': 'object',
    'properties': {
        'version': {
            'description': 'The SemVar version number for the format',
            'type': 'string',
            'enum': ['0.1.0'],
        },
        'general': {
            'type': 'object',
            'properties': {
                'interest_rate': {
                    'type': 'number',
                    'minimum': 0,
                },
                'constraints': {
                    'description': 'A list of extra constraints',
                    'type': 'array',
                    'items': {
                        'type': 'string',
                    },
                },
            },
            'required': ['interest_rate'],
            'additionalProperties': False,
        },
        'capacities': {
            'description': 'A list of capacities a converter or storage can '
                           'have',
            'type': 'array',
            'items': {
                'description': 'A specific capacity',
                'type': 'object',
                'properties': {
                    'name': {
                        'type': 'string',
                    },
                    'units': {
                        'type': 'string',
                    },
                    'type': {
                        'type': 'string',
                        'enum': ['Binary', 'Integer', 'List', 'Continuous'],
                    },
                    'options': {
                        'anyOf': [
                            {
                                'type': 'null',
                            },
                            {
                                'type': 'array',
                                'items': {
                                    'type': 'number',
                                },
                            },
                        ],
                    },
                    'bounds': {
                        'type': 'object',
                        'properties': {
                            'lower': {
                                'type': 'number',
                            },
                            'upper': {
                                'type': 'number',
                            },
                        },
                        'additionalProperties': False,
                    },
                },
                'required': ['name', 'type'],
                'additionalProperties': False,
            },
        },
        'streams': {
            'type': 'array',
            'items': {
                'type': 'object',
                'properties': {
                    'name': {
                        'type': 'string',
                    },
                    'availability': {
                        'type': 'string',
                    },
                    'price': {
                        'anyOf': [
                            {
                                'type': 'number',
                                'minimum': 0,
                            },
                            {
                                'description': 'A reference to a time series',
                                'type': 'string',
                            },
                        ],
                    },
                    'export_price': {
                        'type': 'number',
                        'minimum': 0,
                    },
                    'co2': {
                        'type': 'number',
                        'minimum': 0,
                    },
                },
                'required': ['name', 'price', 'export_price', 'co2'],
                'additionalProperties': False,
            },
        },
        'converters': {
            'type': 'array',
            'items': {
                'type': 'object',
                'properties': {
                    'name': {
                        'type': 'string',
                    },
                    'capacity': {
                        'anyOf': [
                            {
                                'description': 'A reference to a capacity',
                                'type': 'string',
                                'default': '',
                            },
                            {
                                'type': 'number',
                                'minimum': 0,
                            },
                        ],
                    },
                    'fixed_capital_cost': {
                        'type': 'number',
                        'minimum': 0,
                        'default': 0,
                    },
                    'capital_cost': {
                        'type': 'number',
                        'minimum': 0,
                        'default': 0,
                    },
                    'annual_maintenance_cost': {
                        'type': 'number',
                        'minimum': 0,
                    },
                    'usage_maintenance_cost': {
                        'type': 'number',
                    },
                    'efficiency': {
                        'type': 'number',
                    },
                    'lifetime': {
                        'type': 'number',
                    },
                    'output_ratio': {
                        'type': 'number',
                    },
                    'min_load': {
                        'type': 'number',
                    },
                    'inputs': {
                        'type': 'array',
                        'items': {
                            'type': 'string',
                        },
                    },
                    'outputs': {
                        'type': 'array',
                        'items': {
                            'type': 'string',
                        },
                    },
                },
                'additionalProperties': False,
            },
        },
        'storages': {
            'type': 'array',
            'items': {
                'type': 'object',
                'properties': {
                    'name': {
                        'type': 'string',
                    },
                    'stream': {
                        'type': 'string',
                    },
                    'capacity': {
                        'anyOf': [
                            {
                                'description': 'A reference to a capacity',
                                'type': 'string',
                            },
                            {
                                'type': 'number',
                                'minimum': 0,
                            },
                        ],
                    },
                    'cost': {
                        'type': 'number',
                        'minimum': 0,
                    },
                    'lifetime': {
                        'type': 'number',
                        'minimum': 0,
                    },
                    'charge_efficiency': {
                        'type': 'number',
                    },
                    'discharge_efficiency': {
                        'type': 'number',
                    },
                    'decay': {
                        'type': 'number',
                    },
                    'max_charge': {
                        'type': 'number',
                    },
                    'max_discharge': {
                        'type': 'number',
                    },
                    'min_state': {
                        'type': 'number',
                    },
                },
                'additionalProperties': False,
            },
        },
        'system_types': {
            'type': 'array',
            'items': {
                'type': 'object',
                'properties': {
                    'name': {
                        'type': 'string',
                    },
                    'technologies': {
                        'type': 'array',
                        'items': {
                            'type': 'string',
                        },
                    },
                },
                'additionalProperties': False,
            },
        },
        'time_series': {
            'type': 'array',
            'items': {
                'type': 'object',
                'properties': {
                    'id': {
                        'type': 'string',
                    },
                    'type': {
                        'type': 'string',
                        'enum': ['Demand', 'Source', 'Price'],
                    },
                    'stream': {
                        'type': 'string',
                    },
                    'node': {
                        'anyOf': [
                            {
                                'description': 'A reference to a network node',
                                'type': 'integer',
                            },
                            {
                                'type': 'null',
                            },
                        ],
                    },
                    'units': {
                        'type': 'string',
                    },
                    'source': {
                        'type': 'null',
                    },
                    'data': {
                        'type': 'array',
                        'items': {
                            'type': 'number',
                        },
                    },
                },
                'additionalProperties': False,
            },
        },
        'network': {
            'type': 'object',
            'properties': {
                'nodes': {
                    'type': 'array',
                    'items': {
                        'type': 'object',
                        'properties': {
                            'id': {
                                'type': 'number',
                            },
                            'coords': {
                                'type': 'object',
                                'properties': {
                                    'latitude': {
                                        'type': 'number',
                                    },
                                    'longitude': {
                                        'type': 'number',
                                    },
                                    'x': {
                                        'type': 'number',
                                    },
                                    'y': {
                                        'type': 'number',
                                    },
                                },
                                'dependencies': {
                                    'latitude': ['longitude'],
                                    'longitude': ['latitude'],
                                    'x': ['y'],
                                    'y': ['x'],
                                },
                                'additionalProperties': False,
                            },
                            'building': {
                                'type': 'object',
                                'properties': {
                                    'id': {
                                        'type': 'integer',
                                    },
                                    'type': {
                                        'type': 'string',
                                    },
                                },
                                'additionalProperties': False,
                            },
                            'system': {
                                'type': 'object',
                                'properties': {
                                    'id': {
                                        'type': 'integer',
                                    },
                                    'type': {
                                        'description': 'A reference to a '
                                                       'system in the system '
                                                       'types',
                                        'type': 'string',
                                    },
                                },
                                'additionalProperties': False,
                            },
                        },
                        'additionalProperties': False,
                    },
                },
                'links': {
                    'type': 'array',
                    'items': {
                        'type': 'object',
                        'properties': {
                            'id': {
                                'type': 'integer',
                            },
                            'start_id': {
                                'description': 'A reference to a network node',
                                'type': 'integer',
                            },
                            'end_id': {
                                'description': 'A reference to a network node',
                                'type': 'integer',
                            },
                            'type': {
                                'type': 'string',
                            },
                            'length': {
                                'type': 'number',
                            },
                            'capacity': {
                                'type': 'number',
                            },
                            'voltage': {
                                'type': 'number',
                            },
                            'electrical_resistance': {
                                'type': 'number',
                            },
                            'electrical_reactance': {
                                'type': 'number',
                            },
                            'total_thermal_loss': {
                                'type': 'number',
                            },
                            'total_pressure_loss': {
                                'type': 'number',
                            },
                            'operating_temperature': {
                                'type': 'number',
                            },
                        },
                        'additionalProperties': False,
                    },
                },
            },
            'required': ['nodes', 'links'],
            'additionalProperties': False,
        },
    },
    'required': ['version', 'general', 'capacities', 'streams', 'converters',
                 'storages', 'system_types', 'time_series'],
    'additionalProperties': False,
}


class RequestValidationError(Exception):
    """The request format instance failed to validate against the SCHEMA."""


def validate(instance: dict) -> None:
    """Validate the instance against the schema.

    Args:
        instance: The potential instance of the schema

    Raises:
        ValidationError: the instance does not match the schema
    """
    try:
        jsonschema.validate(instance, SCHEMA)
    except jsonschema.ValidationError as exc:
        raise RequestValidationError from exc
